# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.trial import unittest

from afkak import KafkaClient, Producer
from afkak.common import (
    PRODUCER_ACK_ALL_REPLICAS, FailedPayloadsError, FetchRequest,
    KafkaUnavailableError, NotLeaderForPartitionError, RequestTimedOutError,
    TopicAndPartition, UnknownTopicOrPartitionError,
)
from afkak.test.int.intutil import (
    IntegrationMixin, ensure_topic_creation, kafka_versions,
)
from afkak.test.testutil import async_delay, random_string

log = logging.getLogger(__name__)


class TestFailover(IntegrationMixin, unittest.TestCase):
    harness_kw = dict(
        replicas=3,
        partitions=7,
    )
    # We want a short timeout on message sending for this test, since
    # we are expecting failures when we take down the brokers
    client_kw = dict(timeout=1000)

    @kafka_versions("all")
    @inlineCallbacks
    def test_switch_leader(self):
        """
        Produce messages while killing the coordinator broker.

        Note that in order to avoid loss of acknowledged writes the producer
        must request acks of -1 (`afkak.common.PRODUCER_ACK_ALL_REPLICAS`).
        """
        producer = Producer(
            self.client,
            req_acks=PRODUCER_ACK_ALL_REPLICAS,
            max_req_attempts=100,
        )
        topic = self.topic
        try:
            for index in range(1, 3):
                # cause the client to establish connections to all the brokers
                log.debug("Pass: %d. Sending 10 random messages", index)
                yield self._send_random_messages(producer, topic, 10)

                # kill leader for partition 0
                log.debug("Killing leader of partition 0")
                broker, kill_time = self._kill_leader(topic, 0)

                log.debug("Sending 1 more message: 'part 1'")
                yield producer.send_messages(topic, msgs=[b'part 1'])
                log.debug("Sending 1 more message: 'part 2'")
                yield producer.send_messages(topic, msgs=[b'part 2'])

                # send to new leader
                log.debug("Sending 10 more messages")
                yield self._send_random_messages(producer, topic, 10)

                # Make sure the ZK ephemeral time (~6 seconds) has elapsed
                wait_time = (kill_time + 6.5) - time.time()
                if wait_time > 0:
                    log.debug("Waiting: %4.2f for ZK timeout", wait_time)
                    yield async_delay(wait_time)
                # restart the kafka broker
                log.debug("Restarting leader broker %r", broker)
                broker.restart()

                # count number of messages
                log.debug("Getting message count")
                count = yield self._count_messages(topic)
                self.assertGreaterEqual(count, 22 * index)
        finally:
            log.debug("Stopping the producer")
            yield producer.stop()
            log.debug("Producer stopped")

        log.debug("Test complete.")

    @inlineCallbacks
    def _send_random_messages(self, producer, topic, n):
        for _j in range(n):
            resp = yield producer.send_messages(
                topic, msgs=[random_string(10).encode()])

            self.assertFalse(isinstance(resp, Exception))

            if resp:
                self.assertEqual(resp.error, 0)

    def _kill_leader(self, topic, partition):
        leader = self.client.topics_to_brokers[
            TopicAndPartition(topic, partition)]
        broker = self.harness.brokers[leader.node_id]
        broker.stop()
        return (broker, time.time())

    @inlineCallbacks
    def _count_messages(self, topic):
        log.debug("Counting messages on topic %s", topic)
        messages = []
        client = KafkaClient(self.harness.bootstrap_hosts,
                             clientId="CountMessages", timeout=500,
                             reactor=self.reactor)

        try:
            yield ensure_topic_creation(client, topic, fully_replicated=False)

            # Need to retry this until we have a leader...
            while True:
                # Ask the client to load the latest metadata. This may avoid a
                # NotLeaderForPartitionError I was seeing upon re-start of the
                # broker.
                yield client.load_metadata_for_topics(topic)
                # if there is an error on the metadata for the topic, wait
                errno = client.metadata_error_for_topic(topic)
                if errno == 0:
                    break
                else:
                    log.debug("Topic %s in error errno=%d", topic, errno)
                    yield async_delay(1.0)

            # Ok, should be safe to get the partitions now...
            partitions = client.topic_partitions[topic]

            requests = [FetchRequest(topic, part, 0, 1024 * 1024)
                        for part in partitions]
            resps = []
            while not resps:
                try:
                    log.debug("_count_message: Fetching messages")
                    resps = yield client.send_fetch_request(requests, max_wait_time=400)
                except (NotLeaderForPartitionError,
                        UnknownTopicOrPartitionError,
                        KafkaUnavailableError):  # pragma: no cover
                    log.debug("_count_message: Metadata err, retrying...")
                    yield client.load_metadata_for_topics(topic)
                except FailedPayloadsError as e:  # pragma: no cover
                    if not e.args[1][0][1].check(RequestTimedOutError):
                        raise
                    log.debug("_count_message: Timed out err, retrying...")
        finally:
            yield client.close()
        for fetch_resp in resps:
            messages.extend(list(fetch_resp.messages))

        log.debug("Got %d messages: %r", len(messages), messages)

        returnValue(len(messages))



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Cyan, Inc.
# Copyright (C) 2016-2019 Ciena Corporation

from setuptools import find_packages, setup

# NB: This version is extracted by the Makefile using awk; don't change the
# formatting here!
version = "19.10.0"

with open('README.md', 'r') as fin:
    readme_lines = fin.readlines()
long_description = ''.join(readme_lines[
    readme_lines.index('<!-- LONG_DESCRIPTION_START -->\n') + 1:
    readme_lines.index('<!-- LONG_DESCRIPTION_END -->\n'):
])

setup(
    name="afkak",
    version=version,
    install_requires=[
        'attrs >= 19.2.0',  # For functioning auto_exc=True.
        'six',
        'Twisted >= 18.7.0',  # First release with @inlineCallbacks cancellation.
    ],
    # Afkak requires both b'' and u'' syntax, so it isn't compatible with early
    # Python 3 releases. Additionally, Python 3.3 is not supported because
    # nobody uses it, though we don't forbid install on that version.
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    extras_require={
        'FastMurmur2': ['pyhash'],
        'snappy': ['python-snappy>=0.5'],
    },

    packages=find_packages(),
    include_package_data=True,

    zip_safe=True,

    author="Robert Thille",
    author_email="rthille@ciena.com",
    maintainer="Tom Most",
    maintainer_email="twm@freecog.net",
    url="https://github.com/ciena/afkak",
    project_urls={
        'Documentation': 'https://afkak.readthedocs.io/en/latest/',
        'Source': 'https://github.com/ciena/afkak',
        'Issues': 'https://github.com/ciena/afkak/issues',
    },
    license="Apache License 2.0",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Communications',
        'Topic :: System :: Distributed Computing',
    ],

    description="Twisted Python client for Apache Kafka",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Kafka client', 'distributed messaging', 'txkafka'],
)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2016, 2017, 2018, 2019 Ciena Corporation

from __future__ import absolute_import

import logging
from collections import defaultdict
from numbers import Integral

from twisted.internet.defer import CancelledError as tid_CancelledError
from twisted.internet.defer import (
    Deferred, DeferredList, fail, inlineCallbacks, returnValue, succeed,
)
from twisted.internet.task import LoopingCall
from twisted.python.failure import Failure

from ._util import _coerce_topic
from .common import (
    CODEC_NONE, PRODUCER_ACK_LOCAL_WRITE, PRODUCER_ACK_NOT_REQUIRED,
    BrokerResponseError, CancelledError, FailedPayloadsError, KafkaError,
    NoResponseError, NotLeaderForPartitionError, ProduceRequest, SendRequest,
    TopicAndPartition, UnknownTopicOrPartitionError, UnsupportedCodecError,
)
from .kafkacodec import _SUPPORTED_CODECS, create_message_set
from .partitioner import RoundRobinPartitioner

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

BATCH_SEND_SECS_COUNT = 30  # Seconds
BATCH_SEND_MSG_COUNT = 10  # Messages
BATCH_SEND_MSG_BYTES = 32 * 1024  # 32 KBytes


class Producer(object):
    """Write messages to Kafka with retries and batching

    :param client: `KafkaClient` instance to use

    :param partitioner_class:
        Factory for topic partitioners, a callable that accepts a topic and
        list of partition numbers. The default is
        :class:`~afkak.partitioner.RoundRobinPartitioner`.

    :param int req_acks:
        A value indicating the acknowledgements that the server must
        receive before responding to the request

    :param float ack_timeout:
        Value (in milliseconds) indicating a how long the server can wait for
        the above acknowledgements.

    :param int max_req_attempts:
        Number of times we will retry a request to Kafka before failing the
        request.

    :param float retry_interval:
        Initial retry interval in seconds, defaults to INIT_RETRY_INTERVAL.

    :param codec:
        Compression codec to apply to messages. Default: `CODEC_NONE`.

    :param bool batch_send:
        If True, messages are sent in batches.

    :param int batch_every_n:
        If set, messages are sent in batches of this many messages.

    :param int batch_every_b:
        If set, messages are sent when this many bytes of messages are waiting
        to be sent.

    :param int batch_every_t:
        If set, messages are sent after this many seconds (even if waiting for
        other conditions to apply).  This caps the latency automatic batching
        incurs.
    """

    DEFAULT_ACK_TIMEOUT = 1000  # How long the server should wait (msec)
    DEFAULT_REQ_ATTEMPTS = 10  # Send request up to 10 times before failing
    INIT_RETRY_INTERVAL = 0.25  # Initial retry interval in seconds
    RETRY_INTERVAL_FACTOR = 1.20205  # Factor by which we increase our delay

    _sendLooper = None
    _sendLooperD = None

    def __init__(self, client,
                 partitioner_class=RoundRobinPartitioner,
                 req_acks=PRODUCER_ACK_LOCAL_WRITE,
                 ack_timeout=DEFAULT_ACK_TIMEOUT,
                 max_req_attempts=DEFAULT_REQ_ATTEMPTS,
                 retry_interval=INIT_RETRY_INTERVAL,
                 codec=None,
                 batch_send=False,
                 batch_every_n=BATCH_SEND_MSG_COUNT,
                 batch_every_b=BATCH_SEND_MSG_BYTES,
                 batch_every_t=BATCH_SEND_SECS_COUNT):

        # When messages are sent, the partition of the message is picked
        # by the partitioner object for that topic. The partitioners are
        # created as needed from the "partitioner_class" class and stored
        # by topic in self.partitioners
        self.partitioner_class = partitioner_class
        self.partitioners = {}

        # Set our client, our acks/timeouts, our clock & interval
        self.client = client
        self.req_acks = req_acks
        self.ack_timeout = ack_timeout
        self._max_attempts = max_req_attempts
        self._req_attempts = 0
        self._retry_interval = self._init_retry_interval = retry_interval

        # For efficiency, the producer can be set to send messages in
        # batches. In that case, the producer will wait until at least
        # batch_every_n messages are waiting to be sent, or batch_every_b
        # bytes of messages are waiting to be sent, or it has been
        # batch_every_t seconds since the last send
        if not batch_send:
            self.batchDesc = "Unbatched"
            self.batch_every_n = 1
            self.batch_every_b = 1
            self.batch_every_t = None
        else:
            if not isinstance(batch_every_n, Integral):
                msg = "batch_every_n: {0!r} unsupported".format(batch_every_n)
                raise TypeError(msg)
            if not isinstance(batch_every_b, Integral):
                msg = "batch_every_b: {0!r} unsupported".format(batch_every_b)
                raise TypeError(msg)
            self.batch_every_n = batch_every_n
            self.batch_every_b = batch_every_b
            self.batch_every_t = batch_every_t
            self.batchDesc = "{}cnt/{}bytes/{}secs".format(
                batch_every_n, batch_every_b, batch_every_t)
            if batch_every_t:
                self._sendLooper = LoopingCall(self._send_batch)
                self._sendLooper.clock = self.client.reactor
                self._sendLooperD = self._sendLooper.start(
                    batch_every_t, now=False)
                self._sendLooperD.addCallbacks(self._send_timer_stopped,
                                               self._send_timer_failed)

        # Current batch reqs & msgs/bytes, and all outstanding reqs
        self._batch_reqs = []  # Current batch (possibly of 1 for unbatched)
        self._waitingMsgCount = 0
        self._waitingByteCount = 0
        self._outstanding = []  # All currently outstanding requests
        self._batch_send_d = None  # Outstanding client request to send msgs

        # Are we compressing messages, or just sending 'raw'?
        if codec is None:
            codec = CODEC_NONE
        elif codec not in _SUPPORTED_CODECS:
            if not isinstance(codec, Integral):
                raise TypeError("Codec: %r unsupported" % codec)
            raise UnsupportedCodecError("Codec 0x%02x unsupported" % codec)
        self.codec = codec

    def __repr__(self):
        return '<Producer {}:{}:{}:{}>'.format(self.partitioner_class,
                                               self.batchDesc, self.req_acks,
                                               self.ack_timeout)

    def send_messages(self, topic, key=None, msgs=()):
        """
        Given a topic, and optional key (for partitioning) and a list of
        messages, send them to Kafka, either immediately, or when a batch is
        ready, depending on the Producer's batch settings.

        :param str topic: Kafka topic to send the messages to

        :param str key:
            Message key used to determine the topic partition to which the
            messages will be written. Either `bytes` or `None`.

            `None` means that there is no key, but note that that:

            - Kafka does not permit producing unkeyed messages to a compacted topic.
            - The *partitioner_class* may require a non-`None` key
              (`HashedPartitioner` does so).

        :param list msgs:
            A non-empty sequence of message bytestrings to send. `None`
            indicates a ``null`` message (i.e. a tombstone on a compacted
            topic).

        :returns:
            A :class:`~twisted.internet.defer.Deferred` that fires when the
            messages have been received by the Kafka cluster.

            It will fail with `TypeError` when:

            - *topic* is not text (`str` on Python 3, `str` or `unicode` on Python 2)
            - *key* is not `bytes` or `None`
            - *msgs* is not a sequence of `bytes` or `None`

            It will fail with `ValueError` when *msgs* is empty.
        """
        try:
            topic = _coerce_topic(topic)
            if key is not None and not isinstance(key, bytes):
                raise TypeError('key={!r} must be bytes or None'.format(key))

            if not msgs:
                raise ValueError("msgs must be a non-empty sequence")

            msg_cnt = len(msgs)
            byte_cnt = 0
            for index, m in enumerate(msgs):
                if m is None:
                    continue

                if not isinstance(m, bytes):
                    raise TypeError('Message {} to topic {} ({!r:.100}) has type {}, but must have type {}'.format(
                        index, topic, m, type(m).__name__, type(bytes).__name__))

                byte_cnt += len(m)
        except Exception:
            return fail()

        d = Deferred(self._cancel_send_messages)
        self._batch_reqs.append(SendRequest(topic, key, msgs, d))
        self._waitingMsgCount += msg_cnt
        self._waitingByteCount += byte_cnt

        # Add request to list of outstanding reqs' callback to remove
        self._outstanding.append(d)
        d.addBoth(self._remove_from_outstanding, d)
        # See if we have enough messages in the batch to do a send.
        self._check_send_batch()
        return d

    def stop(self):
        """
        Terminate any outstanding requests.

        :returns: :class:``Deferred` which fires when fully stopped.
        """
        self.stopping = True
        # Cancel any outstanding request to our client
        if self._batch_send_d:
            self._batch_send_d.cancel()
        # Do we have to worry about our looping call?
        if self.batch_every_t is not None:
            # Stop our looping call, and wait for the deferred to be called
            if self._sendLooper is not None:
                self._sendLooper.stop()
        # Make sure requests that wasn't cancelled above are now
        self._cancel_outstanding()
        return self._sendLooperD or succeed(None)

    # # Private Methods # #

    def _send_timer_failed(self, fail):
        """
        Our _send_batch() function called by the LoopingCall failed. Some
        error probably came back from Kafka.

        For now, just log the failure and restart the loop
        """
        log.warning('Batch timer failed: %s. Will restart.', fail.value,
                    exc_info=(fail.type, fail.value, fail.getTracebackObject()))
        self._sendLooperD = self._sendLooper.start(
            self.batch_every_t, now=False)

    def _send_timer_stopped(self, lCall):
        """
        We're shutting down, clean up our looping call...

        :param lCall:
            The looping call that was stopped (same as `_sendLooper`).
        """
        self._sendLooper = None
        self._sendLooperD = None

    @inlineCallbacks
    def _next_partition(self, topic, key=None):
        """get the next partition to which to publish

        Check with our client for the latest partitions for the topic, then
        ask our partitioner for the next partition to which we should publish
        for the give key. If needed, create a new partitioner for the topic.
        """
        # check if the client has metadata for the topic
        while self.client.metadata_error_for_topic(topic):
            # client doesn't have good metadata for topic. ask to fetch...
            # check if we have request attempts left
            if self._req_attempts >= self._max_attempts:
                # No, no attempts left, so raise the error
                # FIXME: This can be deceptive. metadata_error_for_topic()
                # returns the error code for UnknownTopicOrPartitonError when
                # nothing is known about the topic.
                BrokerResponseError.raise_for_errno(
                    self.client.metadata_error_for_topic(topic),
                    "Exhausted attempt quota of {}".format(self._max_attempts),
                )
            yield self.client.load_metadata_for_topics(topic)
            if not self.client.metadata_error_for_topic(topic):
                break
            self._req_attempts += 1
            d = Deferred()
            self.client.reactor.callLater(
                self._retry_interval, d.callback, True)
            self._retry_interval *= self.RETRY_INTERVAL_FACTOR
            yield d

        # Ok, should be safe to get the partitions now...
        partitions = self.client.topic_partitions[topic]
        # Do we have a partitioner for this topic already?
        if topic not in self.partitioners:
            # No, create a new paritioner for topic, partitions
            self.partitioners[topic] = \
                self.partitioner_class(topic, partitions)
        # Lookup the next partition
        partition = self.partitioners[topic].partition(key, partitions)
        returnValue(partition)

    def _send_requests(self, parts_results, requests):
        """Send the requests

        We've determined the partition for each message group in the batch, or
        got errors for them.
        """
        # We use these dictionaries to be able to combine all the messages
        # destined to the same topic/partition into one request
        # the messages & deferreds, both by topic+partition
        reqsByTopicPart = defaultdict(list)
        payloadsByTopicPart = defaultdict(list)
        deferredsByTopicPart = defaultdict(list)

        # We now have a list of (succeeded/failed, partition/None) tuples
        # for the partition lookups we did on each message group, zipped with
        # the requests
        for (success, part_or_failure), req in zip(parts_results, requests):
            if req.deferred.called:
                # Submitter cancelled the request while we were waiting for
                # the topic/partition, skip it
                continue
            if not success:
                # We failed to get a partition for this request, errback to the
                # caller with the failure. Maybe this should retry? However,
                # since this failure is likely to affect an entire Topic, there
                # should be no issues with ordering of messages within a
                # partition of a topic getting out of order. Let the caller
                # retry the particular request if they like, or they could
                # cancel all their outstanding requests in
                req.deferred.errback(part_or_failure)
                continue
            # Ok, we now have a partition for this request, we can add the
            # request for this topic/partition to reqsByTopicPart, and the
            # caller's deferred to deferredsByTopicPart
            topicPart = TopicAndPartition(req.topic, part_or_failure)
            reqsByTopicPart[topicPart].append(req)
            deferredsByTopicPart[topicPart].append(req.deferred)

        # Build list of payloads grouped by topic/partition
        # That is, we bundle all the messages destined for a given
        # topic/partition, even if they were submitted by different
        # requests into a single 'payload', and then we submit all the
        # payloads as a list to the client for sending to the various
        # brokers. The finest granularity of success/failure is at the
        # payload (topic/partition) level.
        payloads = []
        for (topic, partition), reqs in reqsByTopicPart.items():
            msgSet = create_message_set(reqs, self.codec)
            req = ProduceRequest(topic, partition, msgSet)
            topicPart = TopicAndPartition(topic, partition)
            payloads.append(req)
            payloadsByTopicPart[topicPart] = req
        # Make sure we have some payloads to send
        if not payloads:
            return
        # send the request
        d = self.client.send_produce_request(
            payloads, acks=self.req_acks, timeout=self.ack_timeout,
            fail_on_error=False)
        self._req_attempts += 1
        # add our handlers
        d.addBoth(self._handle_send_response, payloadsByTopicPart,
                  deferredsByTopicPart)
        return d

    def _complete_batch_send(self, resp):
        """Complete the processing of our batch send operation

        Clear the deferred tracking our current batch processing
        and reset our retry count and retry interval
        Return none to eat any errors coming from up the deferred chain
        """
        self._batch_send_d = None
        self._req_attempts = 0
        self._retry_interval = self._init_retry_interval
        if isinstance(resp, Failure) and not resp.check(tid_CancelledError,
                                                        CancelledError):
            log.error(
                "Failure detected in _complete_batch_send: %r", resp,
                exc_info=(resp.type, resp.value, resp.getTracebackObject()),
            )
        return

    def _check_send_batch(self, result=None):
        """Check if we have enough messages/bytes to send
        Since this can be called from the callback chain, we
        pass through our first (non-self) arg
        """
        if (
            (self.batch_every_n and self.batch_every_n <= self._waitingMsgCount) or
            (self.batch_every_b and self.batch_every_b <= self._waitingByteCount)
        ):
            self._send_batch()
        return result

    def _send_batch(self):
        """
        Send the waiting messages, if there are any, and we can...

        This is called by our LoopingCall every send_every_t interval, and
        from send_messages everytime we have enough messages to send.
        This is also called from py:method:`send_messages` via
        py:method:`_check_send_batch` if there are enough messages/bytes
        to require a send.
        Note, the send will be delayed (triggered by completion or failure of
        previous) if we are currently trying to complete the last batch send.
        """
        # We can be triggered by the LoopingCall, and have nothing to send...
        # Or, we've got SendRequest(s) to send, but are still processing the
        # previous batch...
        if (not self._batch_reqs) or self._batch_send_d:
            return

        # Save a local copy, and clear the global list & metrics
        requests, self._batch_reqs = self._batch_reqs, []
        self._waitingByteCount = 0
        self._waitingMsgCount = 0

        # Iterate over them, fetching the partition for each message batch
        d_list = []
        for req in requests:
            # For each request, we get the topic & key and use that to lookup
            # the next partition on which we should produce
            d_list.append(self._next_partition(req.topic, req.key))
        d = self._batch_send_d = Deferred()
        # Since DeferredList doesn't propagate cancel() calls to deferreds it
        # might be waiting on for a result, we need to use this structure,
        # rather than just using the DeferredList directly
        d.addCallback(lambda r: DeferredList(d_list, consumeErrors=True))
        d.addCallback(self._send_requests, requests)
        # Once we finish fully processing the current batch, clear the
        # _batch_send_d and check if any more requests piled up when we
        # were busy.
        d.addBoth(self._complete_batch_send)
        d.addBoth(self._check_send_batch)
        # Fire off the callback to start processing...
        d.callback(None)

    def _cancel_send_messages(self, d):
        """Cancel a `send_messages` request
        First check if the request is in a waiting batch, of so, great, remove
        it from the batch. If it's not found, we errback() the deferred and
        the downstream processing steps take care of aborting further
        processing.
        We check if there's a current _batch_send_d to determine where in the
        chain we were (getting partitions, or already sent request to Kafka)
        and errback differently.
        """
        # Is the request in question in an unsent batch?
        for req in self._batch_reqs:
            if req.deferred == d:
                # Found the request, remove it and return.
                msgs = req.messages
                self._waitingMsgCount -= len(msgs)
                for m in (_m for _m in msgs if _m is not None):
                    self._waitingByteCount -= len(m)
                # This _should_ be safe as we abort the iteration upon removal
                self._batch_reqs.remove(req)
                d.errback(CancelledError(request_sent=False))
                return

        # If it wasn't found in the unsent batch. We just rely on the
        # downstream processing of the request to check if the deferred
        # has been called and skip further processing for this request
        # Errback the deferred with whether or not we sent the request
        # to Kafka already
        d.errback(
            CancelledError(request_sent=(self._batch_send_d is not None)))
        return

    def _handle_send_response(self, result, payloadsByTopicPart,
                              deferredsByTopicPart):
        """Handle the response from our client to our send_produce_request

        This is a bit complex. Failures can happen in a few ways:

          1. The client sent an empty list, False, None or some similar thing
             as the result, but we were expecting real responses.
          2. The client had a failure before it even tried sending any requests
             to any brokers.

               a. Kafka error: See if we can retry the whole request
               b. Non-kafka: Figure it's a programming error, fail all deferreds

          3. The client sent all the requests (it's all or none) to the brokers
             but one or more request failed (timed out before receiving a
             response, or the brokerclient threw some sort of exception on send
             In this case, the client throws FailedPayloadsError, and attaches
             the responses (NOTE: some can have errors!), and the payloads
             where the send itself failed to the exception.
          4. The client sent all the requests, all responses were received, but
             the Kafka broker indicated an error with servicing the request on
             some of the responses.
        """

        def _deliver_result(d_list, result=None):
            """Possibly callback each deferred in a list with single result"""
            for d in d_list:
                if not isinstance(d, Deferred):
                    # nested list...
                    _deliver_result(d, result)
                else:
                    # We check d.called since the request could have been
                    # cancelled while we waited for the response
                    if not d.called:
                        d.callback(result)

        def _do_retry(payloads):
            # We use 'fail_on_error=False' because we want our client to
            # process every response that comes back from the brokers so
            # we can determine which requests were successful, and which
            # failed for retry
            d = self.client.send_produce_request(
                payloads, acks=self.req_acks, timeout=self.ack_timeout,
                fail_on_error=False)
            self._req_attempts += 1
            # add our handlers
            d.addBoth(self._handle_send_response, payloadsByTopicPart,
                      deferredsByTopicPart)
            return d

        def _cancel_retry(failure, dc):
            # Cancel the retry callLater and pass-thru the failure
            dc.cancel()
            # cancel all the top-level deferreds associated with the request
            _deliver_result(deferredsByTopicPart.values(), failure)
            return failure

        def _check_retry_payloads(failed_payloads_with_errs):
            """Check our retry count and retry after a delay or errback

            If we have more retries to try, create a deferred that will fire
            with the result of delayed retry. If not, errback the remaining
            deferreds with failure

            Params:
            failed_payloads - list of (payload, failure) tuples
            """
            # Do we have retries left?
            if self._req_attempts >= self._max_attempts:
                # No, no retries left, fail each failed_payload with its
                # associated failure
                for p, f in failed_payloads_with_errs:
                    t_and_p = TopicAndPartition(p.topic, p.partition)
                    _deliver_result(deferredsByTopicPart[t_and_p], f)
                return
            # Retries remain!  Schedule one...
            d = Deferred()
            dc = self.client.reactor.callLater(
                self._retry_interval, d.callback, [p for p, f in
                                                   failed_payloads])
            self._retry_interval *= self.RETRY_INTERVAL_FACTOR
            # Cancel the callLater when request is cancelled before it fires
            d.addErrback(_cancel_retry, dc)

            # Reset the topic metadata for all topics which had failed_requests
            # where the failures were of the kind UnknownTopicOrPartitionError
            # or NotLeaderForPartitionError, since those indicate our client's
            # metadata is out of date.
            reset_topics = set()
            for payload, e in failed_payloads:
                if (isinstance(e, NotLeaderForPartitionError) or
                        isinstance(e, UnknownTopicOrPartitionError)):
                    reset_topics.add(payload.topic)
            if reset_topics:
                self.client.reset_topic_metadata(*reset_topics)

            d.addCallback(_do_retry)
            return d

        # The payloads we need to retry, if we still can..
        failed_payloads = []
        # In the case we are sending requests without requiring acks, the
        # brokerclient will immediately callback() the deferred upon send with
        # None. In that case, we just iterate over all the deferreds in
        # deferredsByTopicPart and callback them with None
        # If we are expecting responses/acks, and we get an empty result, we
        # callback with a Failure of NoResponseError
        if not result:
            # Success, but no results, is that what we're expecting?
            if self.req_acks == PRODUCER_ACK_NOT_REQUIRED:
                result = None
            else:
                # We got no result, but we were expecting one? Fail everything!
                result = Failure(NoResponseError())
            _deliver_result(deferredsByTopicPart.values(), result)
            return
        elif isinstance(result, Failure):
            # Failure!  Was it total, or partial?
            if not result.check(FailedPayloadsError):
                # Total failure of some sort!
                # The client was unable to send the request at all. If it's
                # a KafkaError (probably Leader/Partition unavailable), retry
                if result.check(KafkaError):
                    # Yep, a kafak error. Set failed_payloads, and we'll retry
                    # them all below. Set failure for errback to callers if we
                    # are all out of retries
                    failure, result = result, []  # no succesful results, retry
                    failed_payloads = [(p, failure) for p in
                                       payloadsByTopicPart.values()]
                else:
                    # Was the request cancelled?
                    if not result.check(tid_CancelledError):
                        # Uh Oh, programming error? Log it!
                        log.error("Unexpected failure: %r in "
                                  "_handle_send_response", result)
                    # Cancelled, or programming error, we fail the requests
                    _deliver_result(deferredsByTopicPart.values(), result)
                    return
            else:
                # FailedPayloadsError: This means that some/all of the
                # requests to a/some brokerclients failed to send.
                # Pull the successful responses and the failed_payloads off
                # the exception and handle them below. Preserve the
                # FailedPayloadsError as 'failure'
                failure = result
                result = failure.value.args[0]
                failed_payloads = failure.value.args[1]

        # Do we have results? Iterate over them and if the response indicates
        # success, then callback the associated deferred. If the response
        # indicates an error, then setup that request for retry.
        # NOTE: In this case, each failed_payload get it's own error...
        for res in result:
            t_and_p = TopicAndPartition(res.topic, res.partition)
            try:
                BrokerResponseError.raise_for_errno(res.error, res)
            except BrokerResponseError as e:
                p = payloadsByTopicPart[t_and_p]
                failed_payloads.append((p, e))
            else:
                # Success for this topic/partition
                d_list = deferredsByTopicPart[t_and_p]
                _deliver_result(d_list, res)

        # Were there any failed requests to possibly retry?
        if failed_payloads:
            return _check_retry_payloads(failed_payloads)
        return

    def _remove_from_outstanding(self, result, d):
        """ Remove 'd' from the list of outstanding requests"""
        self._outstanding.remove(d)
        return result

    def _cancel_outstanding(self):
        """Cancel all of our outstanding requests"""
        for d in list(self._outstanding):
            d.addErrback(lambda _: None)  # Eat any uncaught errors
            d.cancel()



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test code for KafkaClient class.
"""

from __future__ import absolute_import, division

import logging
import random
import struct
from itertools import cycle
from pprint import pformat

from mock import ANY, Mock
from twisted.internet.defer import Deferred
from twisted.internet.error import ConnectError, ConnectionLost, UserError
from twisted.internet.task import Clock
from twisted.python.failure import Failure
from twisted.test.proto_helpers import MemoryReactorClock
from twisted.trial import unittest

from .. import KafkaClient
from ..client import _normalize_hosts
from ..client import log as client_log
from ..common import (
    BrokerMetadata, CoordinatorNotAvailable, FailedPayloadsError, FetchRequest,
    FetchResponse, KafkaUnavailableError, LeaderNotAvailableError,
    LeaderUnavailableError, NotCoordinator, NotLeaderForPartitionError,
    OffsetAndMessage, OffsetCommitRequest, OffsetCommitResponse,
    OffsetFetchRequest, OffsetFetchResponse, OffsetRequest, OffsetResponse,
    PartitionMetadata, PartitionUnavailableError, ProduceRequest,
    ProduceResponse, RequestTimedOutError, TopicAndPartition, TopicMetadata,
    UnknownError, UnknownTopicOrPartitionError, _LeaveGroupRequest,
    _LeaveGroupResponse,
)
from ..kafkacodec import KafkaCodec, create_message
from .endpoints import BlackholeEndpoint, Connections, FailureEndpoint
from .logtools import capture_logging
from .testutil import first

log = logging.getLogger(__name__)


def createMetadataResp():
    from .test_kafkacodec import create_encoded_metadata_response
    node_brokers = {
        0: BrokerMetadata(0, "brokers1.afkak.example.com", 1000),
        1: BrokerMetadata(1, "brokers1.afkak.example.com", 1001),
        3: BrokerMetadata(3, "brokers2.afkak.example.com", 1000),
    }

    topic_partitions = {
        "topic1": TopicMetadata(
            'topic1', 0, {
                0: PartitionMetadata("topic1", 0, 0, 1, (0, 2), (2,)),
                1: PartitionMetadata("topic1", 1, 1, 3, (0, 1), (0, 1)),
            },
        ),
        "topic2": TopicMetadata(
            'topic2', 0, {
                0: PartitionMetadata("topic2", 0, 0, 0, (), ()),
            },
        ),
        "topic3": TopicMetadata('topic3', 5, {}),
    }
    return create_encoded_metadata_response(node_brokers, topic_partitions)


def encode_metadata_response(brokers, topics):
    from .test_kafkacodec import create_encoded_metadata_response

    broker_map = {bm.node_id: bm for bm in brokers}
    return create_encoded_metadata_response(broker_map, topics)[4:]  # Slice off correlation ID


def find_coordinator_ok(node_id, host, port):
    """
    Encode a successful FindCoordinatorResponse v0.
    """
    host_bytes = host.encode('ascii')
    return b"".join([  # FindCoordinator Response v0
        struct.pack('>h', 0),                # No error
        struct.pack('>i', node_id),        # Coordinator id
        struct.pack('>h', len(host_bytes)),  # The Coordinator host
        host_bytes,
        struct.pack('>i', port),             # The Coordinator port
    ])


def find_coordinator_error(error):
    """
    Encode an error'd out FindCoordinatorResponse v0.

    :param error: Kafka errno
    """
    return b"".join([  # FindCoordinator Response v0
        struct.pack('>h', error),         # Error Code
        struct.pack('>i', -1),            # Coordinator id
        struct.pack('>h', 0),             # The Coordinator host
        struct.pack('>i', 0),             # The Coordinator port
    ])


class TestKafkaClient(unittest.TestCase):
    maxDiff = None

    many_brokers = [
        BrokerMetadata(node_id, host='broker_{}'.format(node_id), port=node_id)
        for node_id in range(1001, 1011)
    ]

    testMetaData = createMetadataResp()

    def client_with_metadata(self, brokers, topic_partitions=None,
                             topic_metadata=None, disconnect_on_timeout=False):
        """
        :param brokers: Broker metadata to load
        :type brokers: List[BrokerMetadata]

        :param topic_partitions:
            Map of topic names to partition count. Partition numbers from 0..n
            will be assigned in the generated metadata. The replica and ISR
            lists for each partition will be empty.

            Mutually exclusive with *topic_metadata*.
        :type topics: Dict[str, int]

        :param topic_metadata:
            Full topic metadata as required by :func:`encode_metadata_response()`.

            Mutually exlusive with *topic_partitions*.

        :param bool disconnect_on_timeout:
            Passed on to KafkaClient.
        """
        broker_id_seq = cycle(bm.node_id for bm in brokers)

        if topic_metadata is not None:
            assert topic_partitions is None
            topic_map = topic_metadata
        elif topic_partitions is None:
            topic_map = {}
        else:
            topic_map = {}
            for topic, partition_count in topic_partitions.items():
                node_id = next(broker_id_seq)
                topic_map[topic] = TopicMetadata(topic, 0, {
                    p: PartitionMetadata(
                        topic=topic,
                        partition=p,
                        partition_error_code=0,  # no error
                        leader=node_id,
                        replicas=(node_id,),
                        isr=(node_id,),
                    )
                    for p in range(partition_count)
                })

        reactor = Clock()
        connections = Connections()
        client = KafkaClient(hosts='bootstrap:9092', reactor=reactor, endpoint_factory=connections,
                             disconnect_on_timeout=disconnect_on_timeout)

        d = client.load_metadata_for_topics()
        conn = connections.accept('bootstrap')
        connections.flush()
        request = self.successResultOf(conn.server.expectRequest(KafkaCodec.METADATA_KEY, 0, ANY))
        request.respond(encode_metadata_response(brokers, topic_map))
        connections.flush()
        self.assertTrue(self.successResultOf(d))

        return reactor, connections, client

    def test_repr(self):
        c = KafkaClient('1.kafka.internal, 2.kafka.internal:9910', clientId='MyClient')
        self.assertEqual((
            "<KafkaClient clientId=MyClient"
            " hosts=1.kafka.internal:9092 2.kafka.internal:9910"
            " timeout=10.0>"
        ), repr(c))

    def test_client_bad_timeout(self):
        self.assertRaises(Exception, KafkaClient, 'kafka.example.com', clientId='MyClient', timeout="100ms")
        self.assertRaises(TypeError, KafkaClient, 'kafka.example.com', clientId='MyClient', timeout=None)

    def test_client_bad_hosts(self):
        """
        `KafkaClient.__init__` raises an exception when passed an invalid
        *hosts* argument.
        """
        self.assertRaises(Exception, KafkaClient, hosts='foo:notaport')

    def test_update_cluster_hosts(self):
        c = KafkaClient(hosts='www.example.com')
        c.update_cluster_hosts('meep.org')
        self.assertEqual(c._bootstrap_hosts, [('meep.org', 9092)])

    def test_update_cluster_hosts_empty(self):
        """
        Attempting to set empty bootstrap hosts raises an exception. The
        configured hosts don't change.
        """
        c = KafkaClient(hosts=[('kafka.example.com', 1234)])
        self.assertRaises(Exception, c.update_cluster_hosts, ['foo:notaport'])
        self.assertEqual(c._bootstrap_hosts, [('kafka.example.com', 1234)])

    def test_send_broker_unaware_request_bootstrap_fail(self):
        """
        Broker unaware requests fail with `KafkaUnavailableError` when boostrap
        fails.

        This scenario makes two connection attempts in random order, one for
        each configured bootstrap host. Both fail.
        """
        client = KafkaClient(
            hosts=['kafka01:9092', 'kafka02:9092'],
            reactor=MemoryReactorClock(),
            # Every connection attempt will immediately fail due to this
            # endpoint, including attempts to bootstrap.
            endpoint_factory=FailureEndpoint,
        )

        d = client.load_metadata_for_topics('sometopic')

        self.failureResultOf(d).trap(KafkaUnavailableError)

    def test_send_broker_unaware_request_brokers_fail(self):
        """
        Broker unaware requests fail with `KafkaUnavailableError` when none of
        the clients can answer requests and the fallback bootstrap also fails.

        In this scenario five connection attempts occur:

        1. A bootstrap request to get cluster metadata. This succeeds and
           returns topic metadata which references two brokers (we will call
           these "topic brokers").
        2. A request to a topic broker. We fail this by timeout.
        3. A request to the other topic broker. We fail this by timeout.
        4. A fallback bootstrap request to a bootstrap broker. We fail this.
        5. A fallback bootstrap request to the other bootstrap broker. We fail
           this.

        At this point all brokers have been tried so the request fails with
        a `KafkaUnavailableError`.
        """
        from .test_kafkacodec import create_encoded_metadata_response

        reactor = MemoryReactorClock()
        conns = Connections()
        client = KafkaClient(
            hosts='bootstrap',
            timeout=10 * 1000,  # 10 seconds in ms
            reactor=reactor,
            endpoint_factory=conns,
            # We want to hit the request timeout to test failover behavior, so
            # this retry policy returns a delay much longer than the timeout.
            retry_policy=lambda failures: 100.0,
        )
        topic_brokers = [
            BrokerMetadata(1, 'broker1', 9092),
            BrokerMetadata(2, 'broker2', 9092),
        ]

        # Bootstrap the client.
        bootstrap_d = client.load_metadata_for_topics('bootstrap_topic')
        self.assertNoResult(bootstrap_d)
        conn = conns.accept('bootstrap')
        conn.pump.flush()
        request = self.successResultOf(conn.server.requests.get())
        request.respond(create_encoded_metadata_response(
            {b.node_id: b for b in topic_brokers},
            {},  # No topic metadata.
        )[4:])
        conn.pump.flush()
        self.assertTrue(self.successResultOf(bootstrap_d))
        self.assertEqual({1, 2}, set(client._brokers.keys()))

        # Now hit all the failure paths.
        client.update_cluster_hosts('bootstrap1,bootstrap2')
        d = client.load_metadata_for_topics('sometopic')
        self.assertNoResult(d)

        conns.fail('broker?', ConnectError())
        # The broker client will wait and retry until the request times out
        # (probably should change this for unaware requests) so advance the
        # clock up to the timeout.
        reactor.pump([1.0] * 10)
        self.assertNoResult(d)
        conns.accept('broker?').client.connectionLost(Failure(ConnectionLost()))
        reactor.pump([1.0] * 10)  # Again with the timeout.
        self.assertNoResult(d)

        # Now it falls back to the bootstrap hosts.
        conns.accept('bootstrap?').client.connectionLost(Failure(ConnectionLost()))
        self.assertNoResult(d)
        conns.fail('bootstrap?', ConnectError())

        self.failureResultOf(d).check(KafkaUnavailableError)

        self.successResultOf(client.close())

    def test_make_request_to_broker_alerts_when_blocked(self):
        """
        When the reactor is blocked the request timeout warning includes enough
        information to point at that issue: it shows that the timeout delayed
        call was invoked later than scheduled.
        """
        reactor, connections, client = self.client_with_metadata(
            brokers=[BrokerMetadata(0, 'kafka', 9092)],
            topic_partitions={"topic": 1},
        )

        with capture_logging(client_log, logging.WARNING) as records:
            d = client.send_produce_request(
                [ProduceRequest("topic", 0, [create_message(b"a")])],
            )

            reactor.advance(client.timeout + 1)  # fire the timeout warning errback

        [record] = records
        self.assertEqual(record.args, (ANY, client.timeout, client.timeout + 1))

        self.failureResultOf(d)

    def test_make_request_to_broker_min_timeout(self):
        """
        Test that request timeouts can be overridden individually by the
        caller by passing the `min_timeout` argument.
        """
        cbArg = []

        def _recordCallback(res):
            # Record how the deferred returned by our mocked broker is called
            cbArg.append(res)
            return res

        d = Deferred().addBoth(_recordCallback)
        mock_broker = Mock()
        mocked_brokers = {('kafka31', 9092): mock_broker}
        # inject broker side effects
        mock_broker.makeRequest.return_value = d
        mock_broker.cancelRequest.side_effect = \
            lambda rId, reason: d.errback(reason)

        reactor = MemoryReactorClock()
        client = KafkaClient(hosts='kafka31:9092', reactor=reactor, timeout=5000)

        # Alter the client's brokerclient dict to use our mocked broker
        client.clients = mocked_brokers
        client._collect_hosts_d = None
        respD = client._make_request_to_broker(mock_broker, 1, b'request',
                                               min_timeout=10.0)

        # Advance past the client timeout of 5000ms. The request should not
        # time out as the timeout has been overridden.
        reactor.advance(5.0)
        self.assertNoResult(respD)

        # Advance past the configured timeout. Now the request times out.
        reactor.pump([
            4.0,  # Advance to the reactor block warning, clear it.
            1.0,  # Advance to the timeout.
        ])
        self.failureResultOf(respD).check(RequestTimedOutError)

        # The timeout happened at the broker level, meaning it was cleared from
        # internal data structures.
        self.assertIs(None, cbArg[0].check(RequestTimedOutError))

    def test_load_metadata_for_topics(self):
        """
        A broker unaware request succeeds if at least one bootstrap host is
        available.
        """
        reactor = Clock()
        conns = Connections()
        client = KafkaClient(
            hosts='kafka1,kafka2,kafka3',
            reactor=reactor,
            endpoint_factory=conns,
        )

        d = client.load_metadata_for_topics('topic1', 'topic2', 'topic3', 'topic4')

        conns.fail('kafka?', ConnectError())
        conns.accept('kafka?').client.connectionLost(Failure(ConnectionLost()))
        conn = conns.accept('kafka?')
        conn.pump.flush()
        request = self.successResultOf(conn.server.expectRequest(
            api_key=KafkaCodec.METADATA_KEY,
            api_version=0,
            correlation_id=1,
        ))

        self.assertEqual(request.rest, (
            b'\x00\x00\x00\x04'
            b'\x00\x06topic1\x00\x06topic2\x00\x06topic3\x00\x06topic4'
        ))
        request.respond(encode_metadata_response(
            [
                BrokerMetadata(1, 'kafka1', 9092),
                BrokerMetadata(2, 'kafka2', 9092),
                BrokerMetadata(3, 'kafka3', 9092),
            ],
            {
                "topic1": TopicMetadata(
                    'topic1', 0, {
                        0: PartitionMetadata("topic1", 0, 0, 1, (0, 2), (2,)),
                        1: PartitionMetadata("topic1", 1, 1, 3, (0, 1), (0, 1)),
                    },
                ),
                "topic2": TopicMetadata(
                    'topic2', 0, {
                        # A partition with no replicas or ISRs.
                        0: PartitionMetadata("topic2", 0, 0, 2, (), ()),
                    },
                ),
                "topic3": TopicMetadata('topic3', LeaderNotAvailableError.errno, {}),
                # This topic has partitions which lack a leader.
                "topic4": TopicMetadata('topic4', UnknownError.errno, {
                    0: PartitionMetadata('topic4', 0, 0, -1, [], []),
                    1: PartitionMetadata('topic4', 1, 0, -1, [], []),
                }),
            },
        ))
        conn.pump.flush()
        self.assertTrue(self.successResultOf(d))

        self.assertEqual(client._brokers, {
            1: BrokerMetadata(1, 'kafka1', 9092),
            2: BrokerMetadata(2, 'kafka2', 9092),
            3: BrokerMetadata(3, 'kafka3', 9092),
        })
        self.assertEqual(client.topics_to_brokers, {
            TopicAndPartition('topic1', 0): BrokerMetadata(1, 'kafka1', 9092),
            TopicAndPartition('topic1', 1): BrokerMetadata(3, 'kafka3', 9092),
            TopicAndPartition('topic2', 0): BrokerMetadata(2, 'kafka2', 9092),
            TopicAndPartition('topic4', 0): None,
            TopicAndPartition('topic4', 1): None,
        })
        # TODO more assertions about datastructures

    def test_get_leader_for_partitions_loads_metadata(self):
        """
        Topic metadata is loaded if it is not available for the required topic.
        """
        T1 = 'topic_name'
        leader = random.choice(self.many_brokers)
        reactor, connections, client = self.client_with_metadata(self.many_brokers)

        d = client._get_leader_for_partition(T1, 0)
        self.assertNoResult(d)  # Not cached.

        conn = connections.accept('broker_*')
        connections.flush()
        request = self.successResultOf(conn.server.expectRequest(KafkaCodec.METADATA_KEY, 0, ANY))
        request.respond(encode_metadata_response([leader], {
            T1: TopicMetadata(T1, 0, {
                0: PartitionMetadata(T1, 0, 0, leader.node_id, (leader.node_id,), (leader.node_id,)),
            }),
        }))
        connections.flush()

        self.assertEqual(
            {TopicAndPartition(topic=T1, partition=0): leader},
            client.topics_to_brokers,
        )
        self.assertEqual(leader, self.successResultOf(d))

    def test_get_leader_for_unassigned_partitions(self):
        """
        _get_partition_leader fails with PartitionUnavailableError when no
        partition is defined for a topic after attempting a metadata reload.
        """
        topic_metadata = {
            'topic_no_partitions': TopicMetadata('topic_no_partitions', 0, {}),
        }
        reactor, connections, client = self.client_with_metadata(self.many_brokers, topic_metadata=topic_metadata)

        self.assertDictEqual({}, client.topics_to_brokers)

        key = TopicAndPartition('topic_no_partitions', 0)
        eFail = PartitionUnavailableError("{} not available".format(key))

        d = client._get_leader_for_partition('topic_no_partitions', 0)

        # Same old, same old...
        conn = connections.accept('broker_*')
        connections.flush()
        request = self.successResultOf(conn.server.expectRequest(KafkaCodec.METADATA_KEY, 0, ANY))
        request.respond(encode_metadata_response([], topic_metadata))
        connections.flush()

        fail1 = self.failureResultOf(d, PartitionUnavailableError)
        # Make sure the error msg is correct
        self.assertEqual(type(eFail), fail1.type)
        self.assertEqual(eFail.args, fail1.value.args)

    def test_get_leader_returns_none_when_noleader(self):
        """
        _get_leader_for_partition tries to reload the metadata for a topic when
        it doesn't know of a leader. When the fresh data contains no leader it
        returns `None`.
        """
        topic_metadata = {
            'topic_noleader': TopicMetadata('topic_noleader', 0, {
                # -1 means "no leader"
                0: PartitionMetadata('topic_noleader', 0, 0, -1, [], []),
                1: PartitionMetadata('topic_noleader', 1, 0, -1, [], []),
            }),
        }
        reactor, connections, client = self.client_with_metadata(self.many_brokers, topic_metadata=topic_metadata)
        self.assertEqual({
            TopicAndPartition('topic_noleader', 0): None,
            TopicAndPartition('topic_noleader', 1): None,
        }, client.topics_to_brokers)

        d = client._get_leader_for_partition('topic_noleader', 0)

        # Same old, same old...
        conn = connections.accept('broker_*')
        connections.flush()
        request = self.successResultOf(conn.server.expectRequest(KafkaCodec.METADATA_KEY, 0, ANY))
        request.respond(encode_metadata_response([], topic_metadata))
        connections.flush()

        self.assertIs(None, self.successResultOf(d))

    def test_get_leader_cached(self):
        """
        When the metadata for a topic is cached the partition leader is
        returned based on that. No network activity is required.
        """
        broker_0 = self.many_brokers[0].node_id
        broker_1 = self.many_brokers[1].node_id
        reactor, connections, client = self.client_with_metadata(self.many_brokers, topic_metadata={
            'topic': TopicMetadata('topic', 0, {
                0: PartitionMetadata('topic', 0, 0, broker_0, (broker_0, broker_1), (broker_0, broker_1)),
                1: PartitionMetadata('topic', 1, 0, broker_1, (broker_1, broker_0), (broker_1, broker_0)),
            }),
        })

        self.assertEqual(
            self.many_brokers[0],
            self.successResultOf(client._get_leader_for_partition('topic', 0)),
        )
        self.assertEqual(
            self.many_brokers[1],
            self.successResultOf(client._get_leader_for_partition('topic', 1)),
        )

    def test_send_produce_request_raises_when_noleader(self):
        """
        send_produce_request requires a partition leader. When none is found in
        the cached metadata it tries to reload the metadata. If there is still
        no leader, it fails with LeaderUnavailableError.
        """
        topic_metadata = {
            'topic_noleader': TopicMetadata('topic_noleader', 0, {
                # -1 means "no leader"
                0: PartitionMetadata('topic_noleader', 0, 0, -1, [], []),
                1: PartitionMetadata('topic_noleader', 1, 0, -1, [], []),
            }),
        }
        reactor, connections, client = self.client_with_metadata(self.many_brokers, topic_metadata=topic_metadata)

        d = client.send_produce_request(payloads=[
            ProduceRequest("topic_noleader", 0, [create_message(b"a"), create_message(b"b")]),
        ])

        # Return metadata that still has no leader.
        conn = connections.accept('broker_*')
        connections.flush()
        request = self.successResultOf(conn.server.expectRequest(KafkaCodec.METADATA_KEY, 0, ANY))
        request.respond(encode_metadata_response(self.many_brokers, topic_metadata))
        connections.flush()

        self.failureResultOf(d, LeaderUnavailableError)

    def test_get_brokerclient(self):
        """
        _get_brokerclient generates _KafkaBrokerClient objects based on cached
        broker metadata.
        """
        reactor, connections, client = self.client_with_metadata([
            BrokerMetadata(1, 'broker_1', 4567),
            BrokerMetadata(2, 'broker_2', 9092),
            BrokerMetadata(3, 'broker_3', 45678),
        ])

        b1 = client._get_brokerclient(1)
        self.assertEqual('broker_1', b1.host)
        self.assertEqual(4567, b1.port)

        # The same broker client is returned each time.
        b2 = client._get_brokerclient(2)
        self.assertIs(b2, client._get_brokerclient(2))

        # Get another one...
        b3 = client._get_brokerclient(3)
        self.assertEqual(3, b3.node_id)
        self.assertEqual('broker_3', b3.host)
        self.assertEqual(45678, b3.port)

        self.assertIs(b1, client._get_brokerclient(1))

        self.assertRaises(KeyError, client._get_brokerclient, 4)

    def test_reset_topic_metadata(self):
        """
        `reset_topic_metadata()` removes topic metadata from the cache
        stored in these attributes:

        - `topic_partitions` — topic name to list of partition numbers
        - `topics_to_brokers` — TopicPartition tuple to BrokerMetadata
        - `topic_errors` — topic name to errno (or 0)
        """
        # Setup the client with the metadata we want start with
        brokers = [
            BrokerMetadata(node_id=1, host='kafka01', port=9092),
            BrokerMetadata(node_id=2, host='kafka02', port=9092),
        ]
        reactor, connections, client = self.client_with_metadata(
            brokers=brokers,
            topic_metadata={
                "Topic1": TopicMetadata("Topic1", 0, {
                    0: PartitionMetadata(
                        topic="Topic1",
                        partition=0,
                        partition_error_code=0,
                        leader=1,
                        replicas=(1,),
                        isr=(1,),
                    ),
                }),
                "Topic2": TopicMetadata("Topic2", 0, {
                    0: PartitionMetadata(
                        topic="Topic2",
                        partition=0,
                        partition_error_code=0,
                        leader=1,
                        replicas=(1,),
                        isr=(1,),
                    ),
                }),
                "Topic3": TopicMetadata("Topic3", 0, {
                    part: PartitionMetadata(
                        topic="Topic3",
                        partition=part,
                        partition_error_code=0,
                        leader=brokers[part % len(brokers)].node_id,
                        replicas=(1, 2),
                        isr=(1, 2),
                    ) for part in range(4)
                }),
            },
        )

        # Check if we can get the error code.
        self.assertEqual(0, client.metadata_error_for_topic("Topic1"))

        # Reset the client's metadata for Topic2
        client.reset_topic_metadata(u'Topic2')

        # No change to brokers...
        # topics_to_brokers gets every (topic,partition) tuple removed
        # for that topic
        expected_topics_to_brokers = {
            TopicAndPartition('Topic1', 0): brokers[0],
            TopicAndPartition('Topic3', 0): brokers[0],
            TopicAndPartition('Topic3', 1): brokers[1],
            TopicAndPartition('Topic3', 2): brokers[0],
            TopicAndPartition('Topic3', 3): brokers[1],
        }
        self.assertEqual(expected_topics_to_brokers, client.topics_to_brokers)

        # topic_paritions gets topic deleted
        expected_topic_partitions = {'Topic1': [0], 'Topic3': [0, 1, 2, 3]}
        self.assertEqual(expected_topic_partitions, client.topic_partitions)

        expected_topic_errors = {'Topic1': 0, 'Topic3': 0}
        self.assertEqual(expected_topic_errors, client.topic_errors)

        # Resetting an unknown topic has no effect
        client.reset_topic_metadata("bogus")
        self.assertEqual(expected_topics_to_brokers, client.topics_to_brokers)
        self.assertEqual(expected_topic_partitions, client.topic_partitions)
        self.assertEqual(expected_topic_errors, client.topic_errors)

        # Resetting with no topics has no effect (no, this is not consistent
        # with load_metadata_for_topics()...)
        client.reset_topic_metadata()
        self.assertEqual(expected_topics_to_brokers, client.topics_to_brokers)
        self.assertEqual(expected_topic_partitions, client.topic_partitions)
        self.assertEqual(expected_topic_errors, client.topic_errors)

    def test_reset_topic_metadata_errors(self):
        """
        `reset_topic_metadata()` clears any topic error code. This is reflected
        in the `topic_errors` dict and the return value of
        `metadata_error_for_topic()`.
        """
        brokers = [
            BrokerMetadata(node_id=1, host='kafka01', port=9092),
            BrokerMetadata(node_id=2, host='kafka02', port=9092),
        ]
        reactor, connections, client = self.client_with_metadata(
            brokers=brokers,
            topic_metadata={
                "Topic2": TopicMetadata("Topic2", 0, {
                    0: PartitionMetadata(
                        topic="Topic2",
                        partition=0,
                        partition_error_code=0,
                        leader=2,
                        replicas=(1, 2),
                        isr=(1, 2),
                    ),
                }),
                "Topic3": TopicMetadata("Topic3", UnknownTopicOrPartitionError.errno, {}),
                "Topic4": TopicMetadata("Topic4", LeaderNotAvailableError.errno, {
                    0: PartitionMetadata(
                        topic="Topic4",
                        partition=0,
                        partition_error_code=0,
                        leader=1,
                        replicas=(1,),
                        isr=(1,),
                    ),
                }),
                "Topic5": TopicMetadata("Topic5", LeaderNotAvailableError.errno, {}),
            },
        )

        # Setup the client with the metadata we want start with
        log.debug('topic_partitions = %s', pformat(client.topic_partitions))
        log.debug('topics_to_brokers = %s', pformat(client.topics_to_brokers))
        log.debug('topic_errors = %s', pformat(client.topic_errors))

        self.assertEqual({
            u'Topic2': 0,
            u'Topic3': UnknownTopicOrPartitionError.errno,
            u'Topic4': LeaderNotAvailableError.errno,
            u'Topic5': LeaderNotAvailableError.errno,
        }, client.topic_errors)

        self.assertEqual(
            client.metadata_error_for_topic("Topic4"),
            LeaderNotAvailableError.errno,
        )
        client.reset_topic_metadata("Topic4")
        self.assertEqual(
            client.metadata_error_for_topic("Topic4"),
            UnknownTopicOrPartitionError.errno,
        )

        # Topic 5 has no partitions, so it _only_ appears in topic_errors.
        self.assertEqual(
            client.metadata_error_for_topic("Topic5"),
            LeaderNotAvailableError.errno,
        )
        client.reset_topic_metadata("Topic5")
        self.assertEqual(client.metadata_error_for_topic("Topic5"),
                         UnknownTopicOrPartitionError.errno)

        self.assertEqual({
            u'Topic2': 0,
            u'Topic3': UnknownTopicOrPartitionError.errno,
        }, client.topic_errors)

        client.reset_all_metadata()
        self.assertEqual({}, client.topic_errors)

    def test_has_metadata_for_topic(self):
        reactor, connections, client = self.client_with_metadata(
            brokers=[BrokerMetadata(node_id=1, host='kafka01', port=9092)],
            topic_partitions={"Topic1": 1, "Topic2": 1, "Topic3": 4},
        )

        log.debug('topic_partitons = %r', client.topic_partitions)
        self.assertTrue(client.has_metadata_for_topic('Topic1'))
        self.assertTrue(client.has_metadata_for_topic('Topic2'))
        self.assertTrue(client.has_metadata_for_topic('Topic3'))
        self.assertFalse(client.has_metadata_for_topic("Unknown"))

    def test_reset_all_metadata(self):
        brokers = [
            BrokerMetadata(node_id=1, host='kafka01', port=9092),
            BrokerMetadata(node_id=2, host='kafka02', port=9092),
        ]
        reactor, connections, client = self.client_with_metadata(
            brokers, topic_partitions={"Topic1": 1, "Topic2": 1, "Topic3": 4},
        )

        # FIXME: Patch this in the client with the metadata we want start with
        client._group_to_coordinator = {
            u'ConsumerGroup1': brokers[1],
        }

        self.assertTrue(client.has_metadata_for_topic('Topic1'))
        self.assertTrue(client.has_metadata_for_topic('Topic2'))
        self.assertTrue(client.has_metadata_for_topic('Topic3'))
        self.assertFalse(client.has_metadata_for_topic("Unknown"))

        client.reset_all_metadata()

        self.assertEqual(client.topics_to_brokers, {})
        self.assertEqual(client.topic_partitions, {})
        self.assertEqual(client.consumer_group_to_brokers, {})

    def test_client_close(self):
        """
        close() terminates any outstanding broker connections, ignoring any
        errors.
        """
        reactor, connections, client = self.client_with_metadata(brokers=[
            BrokerMetadata(1, 'kafka0', 9092),
            BrokerMetadata(2, 'kafka2', 9092),
        ])

        request_d = client.load_metadata_for_topics()
        conn1 = connections.accept('kafka*')
        reactor.advance(client.timeout)  # Time out first attempt, try second broker.
        self.assertNoResult(request_d)
        conn2 = connections.accept('kafka*')
        conn2.client.transport.disconnectReason = UserError()

        close_d = client.close()
        self.assertNoResult(close_d)  # Waiting for connection shutdown.

        connections.flush()  # Complete connection shutdown.
        self.assertIs(None, self.successResultOf(close_d))
        self.assertTrue(conn1.server.transport.disconnected)
        self.assertTrue(conn2.server.transport.disconnected)
        # FIXME Afkak #71: The bootstrap retry loop leaves delayed calls (for timeouts)
        # in the reactor.
        self.assertEqual([], reactor.getDelayedCalls())

    def test_client_close_no_clients(self):
        client = KafkaClient(
            hosts='kafka',
            reactor=Clock(),
            endpoint_factory=BlackholeEndpoint,
        )
        self.successResultOf(client.close())
        self.assertEqual([], client.reactor.getDelayedCalls())

    def test_client_close_poisons(self):
        """
        Once close() has been called the client cannot be used to make further
        requests.
        """
        reactor, connections, client = self.client_with_metadata(brokers=[
            BrokerMetadata(1, 'kakfa1', 9092),
        ])

        self.successResultOf(client.close())

        self.failureResultOf(client.load_metadata_for_topics())
        # FIXME Afkak #71: Reject new requests with a clean exception. We
        # currently get a KafkaUnavailableError that wraps an AttributeError
        # due to self.clients being None.
        # self.failureResultOf(client.load_metadata_for_topics(), ClientError)

    def test_client_close_during_metadata_load(self):
        reactor = Clock()
        client = KafkaClient(
            reactor=reactor,
            hosts='kafka',
            # Every connection attempt hangs forever.
            endpoint_factory=BlackholeEndpoint,
        )
        load_d = client.load_metadata_for_topics()
        self.assertNoResult(load_d)

        self.assertIs(None, self.successResultOf(client.close()))
        self.assertEqual([], reactor.getDelayedCalls())

        # XXX: This API doesn't make sense. This deferred should fail with an
        # exception like ClientClosed, not succeed as if metadata has loaded.
        # FIXME: Afkak #71: This doesn't actually fail, though it should.
        # self.assertIs(None, self.successResultOf(load_d))

    def test_load_consumer_metadata_for_group(self):
        """
        Test that a subsequent load request for the same group before the
        the previous one has completed does not make a new request, and that
        a request for a different group does make a new request. Also, check
        That once the request completes for 'Group1', that subsequent requests
        will make a new request.
        """
        G1 = u"ConsumerGroup1"
        G2 = u"ConsumerGroup2"
        response = b"".join([  # FindCoordinator Response v0
            struct.pack('>h', 0),                        # Error Code
            struct.pack('>i', 1001),                     # Coordinator id
            struct.pack('>h', len(b"host1")), b"host1",  # The Coordinator host
            struct.pack('>i', 9092),                     # The Coordinator port
        ])
        reactor, connections, client = self.client_with_metadata(brokers=[])

        load1_d = client.load_consumer_metadata_for_group(G1)
        load2_d = client.load_consumer_metadata_for_group(G2)
        load3_d = client.load_consumer_metadata_for_group(G1)
        self.assertNoResult(load1_d)
        self.assertNoResult(load2_d)
        self.assertNoResult(load3_d)

        # Now 'send' a response to the first/3rd requests
        conn = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(response)
        connections.flush()

        # The G1 requests completed and the client's consumer metadata was updated.
        self.assertTrue(self.successResultOf(load1_d))
        self.assertNoResult(load2_d)
        self.assertTrue(self.successResultOf(load3_d))
        self.assertEqual({
            u'ConsumerGroup1': BrokerMetadata(node_id=1001, host='host1', port=9092),
        }, client.consumer_group_to_brokers)

        # Once a response has been received new calls trigger a fresh request.
        load4_d = client.load_consumer_metadata_for_group(G1)
        self.assertNoResult(load4_d)
        # The request goes to the host returned in the first request rather
        # than the boostrap host. This verifies that the broker metatdata was
        # merged into the client cache.
        connections.accept('host1')

    def test_load_consumer_metadata_for_group_failure(self):
        """
        A network-level error when fetching consumer metadata is transformed
        into a NoCoordinator failure.
        """
        G1 = u"ConsumerGroup1"
        response = b"".join([  # FindCoordinator Response v0
            struct.pack('>h', 15),             # Error Code
            struct.pack('>i', -1),             # Coordinator id
            struct.pack('>h', len(b"")), b"",  # The Coordinator host
            struct.pack('>i', -1),             # The Coordinator port
        ])
        reactor, connections, client = self.client_with_metadata(brokers=[])

        load_d = client.load_consumer_metadata_for_group(G1)

        # Now 'send' a response to the request
        conn = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(response)
        connections.flush()

        # The G1 requests completed and the client's consumer metadata was updated.
        self.failureResultOf(load_d, CoordinatorNotAvailable)

    def test_load_consumer_metadata_for_group_unavailable(self):
        """
        A response that indicates no cooordinator is available causes the
        metadata fetch to fail with CoordinatorNotAvailable.
        """
        G1 = u"ConsumerGroup1"
        response = b"".join([  # FindCoordinator Response v0
            struct.pack('>h', CoordinatorNotAvailable.errno),
            struct.pack('>i', -1),             # Coordinator id
            struct.pack('>h', len(b"")), b"",  # The Coordinator host
            struct.pack('>i', -1),             # The Coordinator port
        ])
        reactor, connections, client = self.client_with_metadata(brokers=[])

        load1_d = client.load_consumer_metadata_for_group(G1)
        load2_d = client.load_consumer_metadata_for_group(G1)
        self.assertNoResult(load1_d)
        self.assertNoResult(load2_d)

        # Now 'send' a response to the requests
        conn = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(response)
        connections.flush()

        # The G1 requests completed and the client's consumer metadata was updated.
        self.failureResultOf(load1_d, CoordinatorNotAvailable)
        self.failureResultOf(load2_d, CoordinatorNotAvailable)

    def test_send_produce_request(self):
        """
        send_produce_request routes payloads to the correct Kafka broker. This
        test covers several scenarios:

        1. Regular produce. The deferred fires once the broker responds.
        2. Produce with acks=0. No response is expected, so the deferred fires
           once the request has been written to the OS socket buffer.
        3. A regular produce that results in a NotLeaderForPartiton error. This
           has the side effect of clearing the topic's cached metadata.
        4. Another regular produce. Due to missing metadata a metadata request
           goes out first, then the actual requests.

        This should really be several separate tests.
        """
        T1 = u"Topic1"
        T2 = u"Topic2"

        connections = Connections()
        client = KafkaClient(hosts='kafka31:9092,kafka32:9092',
                             reactor=Clock(),
                             endpoint_factory=connections)

        # Setup the client with the metadata we want it to have
        brokers = [
            BrokerMetadata(node_id=1, host='kafka31', port=9092),
            BrokerMetadata(node_id=2, host='kafka32', port=9092),
        ]
        client._brokers.update({bm.node_id: bm for bm in brokers})
        client.topic_partitions = {
            T1: [0],
            T2: [0],
        }
        client.topics_to_brokers = {
            TopicAndPartition(topic=T1, partition=0): brokers[0],
            TopicAndPartition(topic=T2, partition=0): brokers[1],
        }

        # Setup the payloads
        payloads = [
            ProduceRequest(
                topic=T1,
                partition=0,
                messages=[create_message("{} message {}".format(T1, i).encode('ascii'))
                          for i in range(10)],
            ),
            ProduceRequest(
                topic=T2,
                partition=0,
                messages=[create_message("{} message {}".format(T2, i).encode('ascii'))
                          for i in range(5)],
            ),
        ]

        respD = client.send_produce_request(payloads)

        # Accept the connection attemtps triggered by the request.
        conn31 = connections.accept('kafka31')
        conn32 = connections.accept('kafka32')

        # Dummy up some responses, one from each broker
        resp0 = struct.pack('>ih%dsiihq' % (len(T1)), 1, len(T1), T1.encode(), 1, 0, 0, 10)
        resp1 = struct.pack('>ih%dsiihq' % (len(T2)), 1, len(T2), T2.encode(), 1, 0, 0, 20)

        connections.flush()
        self.successResultOf(conn31.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp0)
        self.successResultOf(conn32.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp1)
        connections.flush()

        # check the results
        results = list(self.successResultOf(respD))
        self.assertEqual(results, [ProduceResponse(T1, 0, 0, 10), ProduceResponse(T2, 0, 0, 20)])

        respD = client.send_produce_request(payloads, acks=0)
        connections.flush()

        # These requests succeed as soon as they are written to a socket because of acks=0.
        self.assertEqual([], list(self.successResultOf(respD)))

        # Read the requests on the server side.
        self.successResultOf(conn31.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY))
        self.successResultOf(conn32.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY))

        # And again, this time with an error coming back...
        respD = client.send_produce_request(payloads)
        connections.flush()

        # Dummy up some responses, one from each broker
        resp0 = struct.pack('>ih%dsiihq' % len(T1), 1, len(T1), T1.encode(), 1, 0, 0, 10)
        resp1 = struct.pack('>ih%dsiihq' % len(T2), 1, len(T2), T2.encode(), 1, 0,
                            NotLeaderForPartitionError.errno, 20)
        self.successResultOf(conn31.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp0)
        self.successResultOf(conn32.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp1)
        connections.flush()

        self.failureResultOf(respD, NotLeaderForPartitionError)
        # *Only* the error'd topic's metadata was reset.
        self.assertTrue(client.has_metadata_for_topic(T1))
        self.assertFalse(client.has_metadata_for_topic(T2))

        # And again, this time with an error coming back...but ignored,
        # and a callback to pre-process the response
        def preprocCB(response):
            return response

        respD = client.send_produce_request(payloads, fail_on_error=False, callback=preprocCB)
        connections.flush()

        # Handle the metadata request. It may have gone to either broker.
        request = self.successResultOf(first([
            conn31.server.requests.get(),
            conn32.server.requests.get(),
        ]))
        request.respond(encode_metadata_response(
            brokers,
            {T2: TopicMetadata(T2, 0, {0: PartitionMetadata(T2, 0, 0, 2, (1,), (1,))})},
        ))
        connections.flush()

        # Metadata has been updated:
        self.assertTrue(client.has_metadata_for_topic(T2))

        # Now the ProduceRequest requests will be sent:
        connections.flush()

        # Dummy up some responses, one from each broker
        resp0 = struct.pack('>ih%dsiihq' % len(T1), 1, len(T1), T1.encode(), 1, 0, 0, 10)
        resp1 = struct.pack('>ih%dsiihq' % len(T2), 1, len(T2), T2.encode(), 1, 0,
                            NotLeaderForPartitionError.errno, 20)
        self.successResultOf(conn31.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp0)
        self.successResultOf(conn32.server.expectRequest(KafkaCodec.PRODUCE_KEY, 0, ANY)).respond(resp1)
        connections.flush()

        results = list(self.successResultOf(respD))
        self.assertEqual(results, [ProduceResponse(T1, 0, 0, 10), ProduceResponse(T2, 0, 6, 20)])

    def test_send_produce_request_timeout(self):
        """
        send_produce_request fails with a FailedPayloadsError when payloads are
        sent, but fail. That failure wraps the details of the underlying
        failures.
        """
        reactor, connections, client = self.client_with_metadata(
            brokers=[BrokerMetadata(0, 'kafka', 9092)],
            topic_partitions={"topic": 1},
        )
        send_payload = ProduceRequest("topic", 0, [create_message(b"a")])

        d = client.send_produce_request([send_payload])
        reactor.advance(client.timeout)  # Time out the request.

        f = self.failureResultOf(d, FailedPayloadsError)
        self.assertEqual([], f.value.responses)
        [[failed_payload, failure]] = f.value.failed_payloads
        self.assertIs(send_payload, failed_payload)
        failure.trap(RequestTimedOutError)

    def test_send_fetch_request(self):
        """
        Test send_fetch_request
        """
        T1 = u"Topic41"
        T2 = u"Topic42"

        # Setup the client with the metadata we want it to have
        brokers = [
            BrokerMetadata(node_id=1, host='kafka41', port=9092),
            BrokerMetadata(node_id=2, host='kafka42', port=9092),
        ]
        reactor, connections, client = self.client_with_metadata(
            brokers=brokers,
            topic_partitions={T1: 1, T2: 1},
        )
        self.assertEqual(client.topic_partitions, {T1: [0], T2: [0]})
        # self.assertEqual(client.topics_to_brokers, {
        #     TopicAndPartition(topic=T1, partition=0): brokers[0],
        #     TopicAndPartition(topic=T2, partition=0): brokers[1],
        # })

        # Setup the payloads
        payloads = [FetchRequest(T1, 0, 0, 1024), FetchRequest(T2, 0, 0, 1024)]

        respD = client.send_fetch_request(payloads)

        # Dummy up some responses, the same from both brokers for simplicity
        msgs = [create_message(m) for m in
                [b"message1", b"hi", b"boo", b"foo", b"so fun!"]]
        ms1 = KafkaCodec._encode_message_set([msgs[0], msgs[1]], offset=0)
        ms2 = KafkaCodec._encode_message_set([msgs[2]], offset=0)
        ms3 = KafkaCodec._encode_message_set([msgs[3], msgs[4]], offset=48)

        # Note this includes a response for a topic/partition we didn't request
        # and further, that response has an error. However,
        # '_send_broker_aware_request' won't return it to us since it doesn't
        # match our request, so we don't error out.
        encoded = struct.pack('>ih%dsiihqi%dsihqi%dsh%dsiihqi%ds' %
                              (len(T1), len(ms1), len(ms2), len(T2), len(ms3)),
                              2,  # Correlation ID, Num Topics
                              len(T1), T1.encode(), 2, 0, 0, 10, len(ms1), ms1, 1,
                              1, 20, len(ms2), ms2,  # Topic41, 2 partitions,
                              # part0, no-err, high-water-mark-offset, msg1
                              # part1, err=1, high-water-mark-offset, msg1
                              len(T2), T2.encode(), 1, 0, 0, 30, len(ms3), ms3)

        # 'send' the responses
        conn41 = connections.accept('kafka41')
        conn42 = connections.accept('kafka42')
        connections.flush()
        self.successResultOf(conn41.server.expectRequest(KafkaCodec.FETCH_KEY, 0, ANY)).respond(encoded)
        self.successResultOf(conn42.server.expectRequest(KafkaCodec.FETCH_KEY, 0, ANY)).respond(encoded)
        connections.flush()

        # check the results
        results = list(self.successResultOf(respD))

        def expand_messages(response):
            return FetchResponse(response.topic, response.partition,
                                 response.error, response.highwaterMark,
                                 list(response.messages))
        expanded_responses = [expand_messages(r) for r in results]
        expect = [FetchResponse(T1, 0, 0, 10, [OffsetAndMessage(0, msgs[0]),
                                               OffsetAndMessage(1, msgs[1])]),
                  FetchResponse(T2, 0, 0, 30, [OffsetAndMessage(48, msgs[3]),
                                               OffsetAndMessage(49, msgs[4])])]
        self.assertEqual(expect, expanded_responses)

        # Again, with a callback
        def preprocCB(response):
            return response

        respD = client.send_fetch_request(payloads, callback=preprocCB)

        # 'send' the responses
        connections.flush()
        self.successResultOf(conn41.server.expectRequest(KafkaCodec.FETCH_KEY, 0, ANY)).respond(encoded)
        self.successResultOf(conn42.server.expectRequest(KafkaCodec.FETCH_KEY, 0, ANY)).respond(encoded)
        connections.flush()
        # check the results
        results = list(self.successResultOf(respD))
        expanded_responses = [expand_messages(r) for r in results]
        expect = [FetchResponse(T1, 0, 0, 10, [OffsetAndMessage(0, msgs[0]),
                                               OffsetAndMessage(1, msgs[1])]),
                  FetchResponse(T2, 0, 0, 30, [OffsetAndMessage(48, msgs[3]),
                                               OffsetAndMessage(49, msgs[4])])]
        self.assertEqual(expect, expanded_responses)

    def test_send_fetch_request_bad_timeout(self):
        client = KafkaClient(hosts='kafka41:9092,kafka42:9092')
        payload = [FetchRequest(u'T1', 0, 0, 1024)]
        timeout = client.timeout * 1000  # client.timeout=secs, max_wait=msecs
        d = client.send_fetch_request(payload, max_wait_time=timeout)
        self.failureResultOf(d, ValueError)

    def test_send_offset_request(self):
        """
        Test send_offset_request
        """
        T1 = "Topic51"
        T2 = "Topic52"

        connections = Connections()
        client = KafkaClient(hosts='kafka51:9092,kafka52:9092',
                             reactor=Clock(),
                             endpoint_factory=connections)

        # Setup the client with the metadata we want it to have
        brokers = [
            BrokerMetadata(node_id=1, host='kafka51', port=9092),
            BrokerMetadata(node_id=2, host='kafka52', port=9092),
        ]
        client._brokers.update({bm.node_id: bm for bm in brokers})
        client.topic_partitions = {
            T1: [0],
            T2: [0],
        }
        client.topics_to_brokers = {
            TopicAndPartition(topic=T1, partition=0): brokers[0],
            TopicAndPartition(topic=T2, partition=0): brokers[1],
        }

        # Setup the payloads
        payloads = [
            OffsetRequest(T1, 0, -1, 20),  # ask for up to 20
            OffsetRequest(T2, 0, -2, 1),  # -2==earliest, only get 1
        ]

        respD = client.send_offset_request(payloads)

        # Dummy up some responses, one from each broker
        resp1 = b"".join([
            # struct.pack(">i", 42),            # Correlation ID
            struct.pack(">i", 1),             # One topics
            struct.pack(">h", len(T1)), T1.encode(),   # First topic
            struct.pack(">i", 1),             # 1 partition

            struct.pack(">i", 0),             # Partition 0
            struct.pack(">h", 0),             # No error
            struct.pack(">i", 3),             # 3 offsets
            struct.pack(">q", 96),            # Offset 96
            struct.pack(">q", 98),            # Offset 98
            struct.pack(">q", 99),            # Offset 99
        ])
        resp2 = b"".join([
            # struct.pack(">i", 68),            # Correlation ID
            struct.pack(">i", 1),             # One topic
            struct.pack(">h", len(T2)), T2.encode(),   # First topic
            struct.pack(">i", 1),             # 1 partition

            struct.pack(">i", 0),             # Partition 0
            struct.pack(">h", 0),             # No error
            struct.pack(">i", 1),             # 1 offset
            struct.pack(">q", 4096),          # Offset 4096
        ])

        # send the responses
        conn51 = connections.accept('kafka51')
        conn52 = connections.accept('kafka52')
        connections.flush()
        self.successResultOf(conn51.server.expectRequest(KafkaCodec.OFFSET_KEY, 0, ANY)).respond(resp1)
        self.successResultOf(conn52.server.expectRequest(KafkaCodec.OFFSET_KEY, 0, ANY)).respond(resp2)
        connections.flush()

        # check the results
        results = list(self.successResultOf(respD))
        self.assertEqual(set(results), set([
            OffsetResponse(topic=T1, partition=0, error=0, offsets=(96, 98, 99)),
            OffsetResponse(topic=T2, partition=0, error=0, offsets=(4096,)),
        ]))

        # Again, with a callback
        def preprocCB(response):
            return response

        respD = client.send_offset_request(payloads, callback=preprocCB)

        connections.flush()
        self.successResultOf(conn51.server.expectRequest(KafkaCodec.OFFSET_KEY, 0, ANY)).respond(resp1)
        self.successResultOf(conn52.server.expectRequest(KafkaCodec.OFFSET_KEY, 0, ANY)).respond(resp2)
        connections.flush()

        # check the results
        results = list(self.successResultOf(respD))
        self.assertEqual(set(results), set([
            OffsetResponse(topic=T1, partition=0, error=0, offsets=(96, 98, 99)),
            OffsetResponse(topic=T2, partition=0, error=0, offsets=(4096,)),
        ]))

    def test_send_offset_fetch_request(self):
        T1 = "Topic71"
        T2 = "Topic72"
        G1 = "ConsumerGroup1"
        reactor, connections, client = self.client_with_metadata(brokers=[])

        # Setup the payloads
        payloads = [OffsetFetchRequest(T1, 71),
                    OffsetFetchRequest(T2, 72),
                    ]

        # Dummy the response
        resp = b"".join([
            struct.pack(">i", 2),             # Two topics
            struct.pack(">h", len(T1)), T1.encode(),   # First topic
            struct.pack(">i", 1),             # 1 partition
            struct.pack(">i", 71),            # Partition 71
            struct.pack(">q", 49),            # Offset 49
            struct.pack(">h", len(b"Metadata1")), b"Metadata1",  # Metadata
            struct.pack(">h", 0),             # No error
            struct.pack(">h", len(T2)), T2.encode(),   # Second topic
            struct.pack(">i", 1),             # 1 partition
            struct.pack(">i", 72),            # Partition 72
            struct.pack(">q", 27),            # Offset 27
            struct.pack(">h", len(b"Metadata2")), b"Metadata2",  # Metadata
            struct.pack(">h", 0),             # No error
        ])

        respD = client.send_offset_fetch_request(G1, payloads)

        # That first lookup for the group should result in one call to
        # load_consumer_metadata_for_group
        conn1 = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn1.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(find_coordinator_ok(node_id=1, host='kafka99', port=9092))
        connections.flush()

        # 'send' the response
        conn2 = connections.accept('kafka99')
        connections.flush()
        req2 = self.successResultOf(conn2.server.expectRequest(KafkaCodec.OFFSET_FETCH_KEY, 1, ANY))
        req2.respond(resp)
        connections.flush()
        # check the results
        results = list(self.successResultOf(respD))
        self.assertEqual(set(results), set([
            OffsetFetchResponse(topic=T1, partition=71, offset=49,
                                metadata=b'Metadata1', error=0),
            OffsetFetchResponse(topic=T2, partition=72, offset=27,
                                metadata=b'Metadata2', error=0),
        ]))

        # Again, with a callback (for full coverage)
        preproc_call_count = [0]

        def preprocCB(response):
            preproc_call_count[0] += 1
            return response

        respD = client.send_offset_fetch_request(G1, payloads,
                                                 callback=preprocCB)

        # 'send' the responses
        connections.flush()
        req2 = self.successResultOf(conn2.server.expectRequest(KafkaCodec.OFFSET_FETCH_KEY, 1, ANY))
        req2.respond(resp)
        connections.flush()
        # check the results
        results = list(self.successResultOf(respD))
        # Test that the preprocessor callback was called (2 responses)
        self.assertEqual(preproc_call_count[0], 2)

        self.assertEqual(set(results), set([
            OffsetFetchResponse(topic=T1, partition=71, offset=49,
                                metadata=b'Metadata1', error=0),
            OffsetFetchResponse(topic=T2, partition=72, offset=27,
                                metadata=b'Metadata2', error=0),
        ]))

    def test_send_offset_fetch_request_failure(self):
        """
        Test that when a request involving a consumer metadata broker fails
        that we reset the cached broker for the consumer group.
        """
        T1 = "Topic71"
        G1 = "ConsumerGroup1"
        reactor, connections, client = self.client_with_metadata(brokers=[])

        # Setup the payload
        payloads = [OffsetFetchRequest(T1, 78)]

        # Dummy the response
        resp = b"".join([
            struct.pack(">i", 1),   # 1 topic
            struct.pack(">h", len(T1)), T1.encode(),  # Topic
            struct.pack(">i", 1),   # 1 partition
            struct.pack(">i", 78),  # Partition 78
            struct.pack(">q", -1),  # Offset -1
            struct.pack(">h", 0),   # Metadata
            struct.pack(">h", NotCoordinator.errno),
        ])

        respD = client.send_offset_fetch_request(G1, payloads)

        # That first lookup for the group should result in one call to
        # load_consumer_metadata_for_group
        conn1 = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn1.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(find_coordinator_ok(node_id=1, host='kafka72', port=9092))
        connections.flush()

        # And the cache of the broker for the consumer group should be set
        self.assertIn(G1, client.consumer_group_to_brokers)

        # 'send' the response
        conn2 = connections.accept('kafka72')
        connections.flush()
        req2 = self.successResultOf(conn2.server.expectRequest(KafkaCodec.OFFSET_FETCH_KEY, 1, ANY))
        req2.respond(resp)
        connections.flush()

        # check the results
        self.assertTrue(self.failureResultOf(respD, NotCoordinator))
        self.assertNotIn(G1, client.consumer_group_to_brokers)
        # cleanup
        client.close()

    def test_send_offset_commit_request(self):
        """test_send_offset_commit_request"""
        T1 = "Topic61"
        T2 = "Topic62"
        G2 = "ConsumerGroup2"
        reactor, connections, client = self.client_with_metadata(brokers=[])

        # Setup the payloads
        respD = client.send_offset_commit_request(G2, payloads=[
            OffsetCommitRequest(T1, 61, 81, -1, b"metadata1"),
            OffsetCommitRequest(T2, 62, 91, -1, b"metadata2"),
        ])

        # The coordinator didn't have any metadata for the group, so it tries to load some.
        conn1 = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn1.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G2.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(find_coordinator_ok(node_id=10, host='kafka123', port=1023))
        connections.flush()

        # Dummy up a response for the commit
        resp1 = b"".join([
            struct.pack(">i", 2),             # Two topics
            struct.pack(">h", len(T1)), T1.encode(),   # First topic
            struct.pack(">i", 1),             # 1 partition
            struct.pack(">i", 61),            # Partition 61
            struct.pack(">h", 0),             # No error
            struct.pack(">h", len(T2)), T2.encode(),   # Second topic
            struct.pack(">i", 1),             # 1 partition
            struct.pack(">i", 62),            # Partition 62
            struct.pack(">h", 0),             # No error
        ])

        # 'send' the responses
        conn2 = connections.accept('kafka123')
        connections.flush()
        req2 = self.successResultOf(conn2.server.expectRequest(KafkaCodec.OFFSET_COMMIT_KEY, 1, ANY))
        req2.respond(resp1)
        connections.flush()

        # check the results
        results = list(self.successResultOf(respD))
        self.assertEqual(set(results), set([
            OffsetCommitResponse(topic=T1, partition=61, error=0),
            OffsetCommitResponse(topic=T2, partition=62, error=0),
        ]))

    def test_send_offset_commit_request_failure(self):
        """
        When the Kafka broker is unavailable, that the proper
        CoordinatorNotAvailable is raised
        """
        T1 = "Topic61"
        G1 = "ConsumerGroup1"
        reactor, connections, client = self.client_with_metadata(brokers=[])

        # Setup the payloads
        d = client.send_offset_commit_request(group=G1, payloads=[
            OffsetCommitRequest(T1, 61, 81, -1, b"metadata1"),
        ])

        # The coordinator didn't have any metadata for the group, so it tries to load some.
        conn1 = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn1.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(find_coordinator_error(CoordinatorNotAvailable.errno))
        connections.flush()

        self.failureResultOf(d, CoordinatorNotAvailable)

    def test_send_request_to_coordinator(self):
        """
        Test the _send_request_to_coordinator method and error handling
        """
        G1 = u"group1"
        reactor, connections, client = self.client_with_metadata(brokers=[])

        # Setup the payloads, encoder & decoder funcs
        d = client._send_request_to_coordinator(
            group=G1,
            payload=_LeaveGroupRequest("group1", "member"),
            encoder_fn=KafkaCodec.encode_leave_group_request,
            decode_fn=KafkaCodec.decode_leave_group_response,
            min_timeout=15,
        )
        # Shouldn't have a result yet. If we do, there was an error
        self.assertNoResult(d)

        # The coordinator didn't have any metadata for the group, so it tries to load some.
        conn1 = connections.accept('bootstrap')
        connections.flush()
        req1 = self.successResultOf(conn1.server.expectRequest(KafkaCodec.FIND_COORDINATOR_KEY, 0, ANY))
        assert G1.encode() in req1.rest  # TODO: Better assert on the request instead of allowing ANY.
        req1.respond(find_coordinator_ok(node_id=1001, host='host1', port=9092))
        connections.flush()

        self.assertEqual({
            u'group1': BrokerMetadata(node_id=1001, host='host1', port=9092),
        }, client.consumer_group_to_brokers)

        conn2 = connections.accept('host1')
        connections.flush()
        req2 = self.successResultOf(conn2.server.expectRequest(KafkaCodec.LEAVE_GROUP_KEY, 0, ANY))
        # dummy response
        req2.respond(struct.pack('>ih', 9, 0))
        connections.flush()

        # check the result. Should be a success response
        self.assertEqual(_LeaveGroupResponse(0), self.successResultOf(d))

    def test_client_disconnect_on_timeout_false(self):
        """
        The connection to a broker is not terminated due to request timeout
        unless `disconnect_on_timeout` is set.
        """
        self._check_disconnect_on_timeout(False)

    def _check_disconnect_on_timeout(self, disconnect_on_timeout):
        reactor, connections, client = self.client_with_metadata(
            brokers=[BrokerMetadata(0, 'kafka', 9092)],
            topic_partitions={"topic": 1},
            disconnect_on_timeout=disconnect_on_timeout,
        )
        payload = ProduceRequest("topic", 0, [create_message(b"a")])

        # Kick off a request to the client to trigger a connection attempt
        respD = client.send_produce_request([payload])

        # Accept the connection attempt, but don't respond.
        conn = connections.accept('kafka')
        connections.flush()

        # Cause that request to timeout
        reactor.advance(client.timeout * 0.91)  # fire the reactor-blocked
        reactor.advance(client.timeout)  # fire the timeout errback
        connections.flush()

        # Check that the request was timed out
        self.failureResultOf(respD)

        # Make sure we didn't try to disconnect
        self.assertEqual(disconnect_on_timeout, conn.client.transport.disconnected)

    def test_client_disconnect_on_timeout_true(self):
        """
        When a client request to a broker times out the connection to that
        broker is terminated.
        """
        self._check_disconnect_on_timeout(True)

    def test_client_topic_fully_replicated(self):
        """test_client_topic_fully_replicated"

        Test the happy path of client.topic_fully_replicated()
        """
        client = KafkaClient(hosts='kafka01:9092,kafka02:9092',
                             reactor=Clock(), endpoint_factory=FailureEndpoint)

        # Setup the client with the metadata we want start with
        client.topic_partitions = {
            'Topic0': [],  # No partitions.
            'Topic2': [0],  # Not in partition_meta below
            'Topic3': [0, 1, 2, 3],  # "Good" partition
        }
        client.partition_meta = {
            TopicAndPartition(topic='Topic3', partition=0):
                PartitionMetadata('Topic3', 0, 0, 1, [1, 2], [1, 2]),
            TopicAndPartition(topic='Topic3', partition=1):
                PartitionMetadata('Topic3', 1, 0, 2, [2, 1], [1, 2]),
            TopicAndPartition(topic='Topic3', partition=2):
                PartitionMetadata('Topic3', 2, 0, 1, [1, 2], [1, 2]),
            TopicAndPartition(topic='Topic3', partition=3):
                PartitionMetadata('Topic3', 3, 0, 2, [2, 1], [1, 2]),
        }

        # Topics which don't have partitions aren't 'fully replicated'
        self.assertFalse(client.topic_fully_replicated('Topic0'))
        # Topics which we don't have any metadata for aren't 'fully replicated'
        self.assertFalse(client.topic_fully_replicated('Topic1'))
        # Topics which are only partially represented aren't 'fully replicated'
        self.assertFalse(client.topic_fully_replicated('Topic2'))

        # Topic with all the partitions having equal replicas & ISRs ARE
        self.assertTrue(client.topic_fully_replicated('Topic3'))


class NormalizeHostsTests(unittest.TestCase):
    def test_bare_host_string(self):
        """
        The default port is inferred when none is given.
        """
        self.assertEqual([('kafka', 9092)], _normalize_hosts('kafka'))
        self.assertEqual([('127.0.0.1', 9092)], _normalize_hosts(u'127.0.0.1'))
        self.assertEqual([('kafka.internal', 9092)], _normalize_hosts('kafka.internal'))
        self.assertEqual([('kafka-1.internal', 9092), ('kafka-2.internal', 9092)],
                         _normalize_hosts(b'kafka-1.internal, kafka-2.internal'))

    def test_string_with_ports(self):
        """
        The string may specify ports.
        """
        self.assertEqual([('kafka', 1234)], _normalize_hosts(u'kafka:1234 '))
        self.assertEqual([('kafka', 1234), ('kafka', 2345)], _normalize_hosts(u'kafka:1234 ,kafka:2345'))
        self.assertEqual([('1.2.3.4', 5555)], _normalize_hosts(b' 1.2.3.4:5555 '))

    def test_sequence(self):
        """
        The input may be a sequence of hostnames or host ports. The default
        port is implied when none is given.
        """
        self.assertEqual(
            [('kafka', 9092)],
            _normalize_hosts([u'kafka', b'kafka', ('kafka', 9092)]),
        )
        self.assertEqual(
            [('kafka1', 9092), ('kafka2', 9092)],
            _normalize_hosts([('kafka2', 9092), ('kafka1', 9092)]),
        )
        self.assertEqual(
            [('kafka1', 1234), ('kafka2', 2345)],
            _normalize_hosts([('kafka2', u'2345'), ('kafka1', b'1234')]),
        )

    def test_ipv6(self):
        """
        IPv6 addresses may be passed as part of a sequence.
        """
        self.assertEqual(
            [('2001:db8::1', 2345), ('::1', 1234)],
            _normalize_hosts([('2001:db8::1', '2345'), (b'::1', 1234)]),
        )



# -*- coding: utf-8 -*-
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import logging
import pprint

from mock import Mock, patch
from twisted.internet import defer, task
from twisted.python.failure import Failure
from twisted.trial import unittest

from afkak import ConsumerGroup
from afkak._group import Coordinator, _ConsumerProtocol, _NeedTopicPartitions
from afkak.common import (
    CoordinatorNotAvailable, IllegalGeneration, InconsistentGroupProtocol,
    InvalidGroupId, NotCoordinator, RebalanceInProgress, RequestTimedOutError,
    RestartError, RestopError, UnknownError, _HeartbeatResponse,
    _JoinGroupRequestProtocol, _JoinGroupResponse, _JoinGroupResponseMember,
    _SyncGroupRequestMember, _SyncGroupResponse,
)
from afkak.kafkacodec import KafkaCodec

log = logging.getLogger(__name__)


def assert_delayed_calls(n, client):
    """
    :param int n: Expected number of delayed calls.
    :param client:
        An object like `afkak.client.KafkaClient` which has a *reactor*
        attribute.
    """
    calls = client.reactor.getDelayedCalls()
    calls_repr = pprint.pformat([vars(c) for c in calls])
    log.debug('assert_delayed_calls(%d) found calls %s', n, calls_repr)
    if len(calls) == n:
        return
    raise AssertionError((
        "Expected {expected} delayed calls at time {now},"
        " but found {actual}:\n{calls}"
    ).format(
        expected=n,
        actual=len(calls),
        now=client.reactor.seconds(),
        calls=calls_repr,
    ))


class Base(unittest.TestCase):
    def mock_client(self, coordinator_responses, topic_partitions=None):
        if topic_partitions is None:
            topic_partitions = {"topic1": [0, 1]}

        def _load_topic_partitions(*topics):
            # NB: Will KeyError for unknown topics, which isn't exactly realistic.
            result = {topic: topic_partitions[topic] for topic in topics}
            return defer.succeed(result)

        client = Mock()
        client.reactor = task.Clock()
        client._send_request_to_coordinator.side_effect = coordinator_responses
        client._load_topic_partitions.side_effect = _load_topic_partitions
        return client

    def make_coordinator(self, client):
        return Coordinator(client, "group_id", ["topic1"])

    def join_response(self, member_id="m1", leader_id="m1"):
        return defer.succeed(_JoinGroupResponse(
            error=0,
            generation_id="g1",
            group_protocol="consumer",
            member_id=member_id,
            leader_id=leader_id,
            members=[
                _JoinGroupResponseMember(
                    member_id,
                    member_metadata=b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x00',
                ),
            ],
        ))

    def sync_response(self, member_id="m1", assignments=None):
        if assignments is None:
            assignments = {"topic1": [0, 1]}  # All partitions per the default topic_partitions
        return defer.succeed(_SyncGroupResponse(
            error=0,
            member_assignment=KafkaCodec.encode_sync_group_member_assignment(
                version=0,
                assignments=assignments,
                user_data=b'',
            ),
        ))


class TestCoordinator(Base):
    def test_send_join_group_request_success(self):
        client = self.mock_client([
            self.join_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.send_join_group_request()
        self.successResultOf(de)
        self.assertEqual(coord.member_id, "m1")
        self.assertEqual(coord.leader_id, "m1")

    def test_send_join_group_request_failure(self):
        client = self.mock_client([
            defer.fail(RebalanceInProgress()),
        ])
        coord = self.make_coordinator(client)
        de = coord.send_join_group_request()
        self.successResultOf(de)
        self.assertEqual(coord.member_id, "")
        self.assertEqual(coord.leader_id, None)
        self.assertEqual(coord._rejoin_needed, True)
        self.assertIn("rejoin_needed", repr(coord))

    def test_send_sync_group_request_success(self):
        client = self.mock_client([
            self.join_response(), self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.send_sync_group_request([])
        self.successResultOf(de)

    def test_send_sync_group_request_failure(self):
        client = self.mock_client([
            defer.fail(RebalanceInProgress()),
        ])
        coord = self.make_coordinator(client)
        de = coord.send_sync_group_request([])
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, True)

    def test_join_no_broker(self):
        """
            fail to retrieve the coordinator broker and retry
        """
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        coord.get_coordinator_broker = Mock(return_value=None)

        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, True)

    def test_join_sync_leader(self):
        """
        Successfully join, assign as leader, and sync
        """
        client = self.mock_client([
            self.join_response(),
            self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, False)
        self.assertEqual(coord.member_id, "m1")
        self.assertEqual(coord.leader_id, "m1")
        self.assertIn("joined", repr(coord))

    def test_join_sync_follower(self):
        """
            Successfully join and sync as follower
        """
        client = self.mock_client([
            self.join_response(leader_id="m2"),
            self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.join_and_sync()
        self.successResultOf(de)

    def test_join_error(self):
        """
            Get an error when joining and retry
        """
        client = self.mock_client([
            defer.fail(RebalanceInProgress()),
            self.join_response(),
            self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, True)
        assert_delayed_calls(1, client)
        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, False)

    def test_sync_error(self):
        """
            Get an error when syncing and retry
        """
        client = self.mock_client([
            self.join_response(),
            defer.fail(RebalanceInProgress()),
            self.join_response(),
            self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, True)
        assert_delayed_calls(1, client)
        de = coord.join_and_sync()
        self.successResultOf(de)
        self.assertEqual(coord._rejoin_needed, False)

    def test_join_fatal_error(self):
        """
            Get an unexpected (non-kafka) error when joining and propagate it
        """
        client = self.mock_client([
            defer.fail(AttributeError()),
        ])
        coord = self.make_coordinator(client)
        start_d = coord._start_d = defer.Deferred()
        de = coord.join_and_sync()
        self.failureResultOf(start_d)
        self.successResultOf(de)

    def test_join_not_needed(self):
        """
            call join_and_sync when not needed
        """
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        _join = coord._join_and_sync = Mock()
        coord._rejoin_d = defer.Deferred()
        coord.join_and_sync()
        _join.assert_not_called()
        coord._rejoin_d = None
        coord._rejoin_needed = False
        coord.join_and_sync()
        _join.assert_not_called()

    def test_join_fatal_exception(self):
        """
            have an exception come out of _join_and_sync
        """
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        coord._start_d = defer.Deferred()
        coord._join_and_sync = Mock(return_value=defer.Deferred())
        result = coord.join_and_sync()
        result.errback(ArithmeticError())
        self.successResultOf(result)

    def test_heartbeat_resync(self):
        """
            run successful heartbeats, get a resync message, and schedule resync
        """
        client = self.mock_client([
            self.join_response(),
            self.sync_response(),
            defer.succeed(_HeartbeatResponse(error=0)),
            defer.fail(RebalanceInProgress()),
            self.join_response(),
            self.sync_response(),
        ])
        coord = self.make_coordinator(client)
        self.successResultOf(coord.join_and_sync())
        self.assertEqual(coord._rejoin_needed, False)
        assert_delayed_calls(1, client)

        # Run the heartbeat, which succeeds.
        client.reactor.advance(5.0)
        self.assertEqual(coord._rejoin_needed, False)
        self.assertFalse(coord._heartbeat_request_d)

        # Run a heartbeat, which fails with RebalanceInProgress
        client.reactor.advance(5.0)
        self.assertEqual(coord._rejoin_needed, True)
        self.assertFalse(coord._heartbeat_request_d)
        # There is a delayed call to join_and_sync():
        assert_delayed_calls(1, client)
        self.assertEqual(coord._rejoin_needed, True)

        client.reactor.advance(0.1)  # Call join_and_sync()
        self.assertEqual(coord._rejoin_needed, False)
        assert_delayed_calls(1, client)  # Heartbeat scheduled.

        self.assertIs(None, coord.join_and_sync())
        self.assertEqual(coord._rejoin_needed, False)

    def test_heartbeat_coordinator_lost(self):
        """
            run a heartbeat that indicates the coordinator has changed
        """
        client = self.mock_client([
            self.join_response(),
            self.sync_response(),
            defer.fail(NotCoordinator()),
        ])
        coord = self.make_coordinator(client)
        self.successResultOf(coord.join_and_sync())

        self.assertEqual(coord._rejoin_needed, False)

        assert_delayed_calls(1, client)
        client.reactor.advance(6.0)
        self.assertFalse(coord._heartbeat_request_d)
        assert_delayed_calls(1, client)
        self.assertEqual(coord._rejoin_needed, True)
        client.reset_consumer_group_metadata.assert_called_with(coord.group_id)

    def test_heartbeat_in_progress(self):
        """
            get a heartbeat timer tick while a request is already in progress
        """
        client = self.mock_client([
            self.join_response(),
            self.sync_response(),
            defer.succeed(_HeartbeatResponse(error=0)),
        ])
        coord = self.make_coordinator(client)
        self.successResultOf(coord.join_and_sync())
        self.assertEqual(coord._rejoin_needed, False)

        assert_delayed_calls(1, client)  # Heartbeat scheduled.

        coord._heartbeat_request_d = defer.Deferred()
        coord.send_heartbeat_request = Mock()
        client.reactor.advance(5.0)
        coord.send_heartbeat_request.assert_not_called()
        coord._heartbeat_request_d = None
        coord._stopping = True
        coord._heartbeat()
        coord.send_heartbeat_request.assert_not_called()
        coord._stopping = False
        coord._heartbeat()
        coord.send_heartbeat_request.assert_any_call()

    def test_leave(self):
        """
            send a leavegroup message
        """
        client = self.mock_client([
            defer.succeed(Mock(error_code=0)),
        ])
        coord = self.make_coordinator(client)
        coord.coordinator_broker = Mock()
        coord.member_id = "m1"
        coord.generation_id = "g1"
        de = coord.send_leave_group_request()
        self.successResultOf(de)
        self.assertEqual(coord.member_id, "")
        self.assertIsNone(coord.generation_id)

    def test_start_stop(self):
        client = self.mock_client([
            self.join_response(), self.sync_response(),
        ])
        coord = self.make_coordinator(client)

        start_d = coord.start()
        self.assertNoResult(start_d)
        self.assertRaises(RestartError, coord.start)

        heatbeat_req_d = coord._heartbeat_request_d = defer.Deferred()
        rejoin_wait_d = coord._rejoin_wait_dc = defer.Deferred()
        rejoin_d = coord._rejoin_d = defer.Deferred()
        stop_d = coord.stop()
        self.successResultOf(stop_d)
        self.successResultOf(start_d)
        self.assertEqual(coord.member_id, "")
        self.assertIsNone(coord.protocol)

        self.failureResultOf(heatbeat_req_d)
        self.failureResultOf(rejoin_wait_d)
        self.failureResultOf(rejoin_d)

    def test_stop_not_started(self):
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        stop = coord.stop()
        self.failureResultOf(stop, RestopError)

    def test_double_stop(self):
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        coord._start_d = defer.Deferred()
        coord._stopping = True
        double_stop = coord.stop()
        self.failureResultOf(double_stop, RestopError)

    def test_get_coordinator_success(self):
        """
        retrieve a coordinator
        """
        client = self.mock_client([
            defer.succeed(Mock(error_code=0)),
        ])
        coord = self.make_coordinator(client)
        client._get_coordinator_for_group.return_value = defer.succeed(None)
        client._get_brokerclient.return_value = None
        de = coord.get_coordinator_broker()
        self.successResultOf(de)
        self.assertEqual(1, len(client.reactor.getDelayedCalls()))

        client._get_coordinator_for_group.return_value = defer.succeed(Mock())
        client._get_brokerclient.return_value = coord
        client.load_metadata_for_topics.return_value = defer.succeed(None)
        de = coord.get_coordinator_broker()
        result = self.successResultOf(de)
        self.assertEqual(result, coord)

    def test_get_coordinator_retry(self):
        """
        fail to retrieve a coordinator and retry
        """
        client = self.mock_client([
            defer.succeed(Mock(error_code=0)),
        ])
        coord = self.make_coordinator(client)
        client._get_coordinator_for_group.return_value = defer.fail(
            RequestTimedOutError())
        client._get_brokerclient.return_value = None
        de = coord.get_coordinator_broker()
        self.successResultOf(de)
        assert_delayed_calls(1, client)  # Heartbeat scheduled.
        client._get_coordinator_for_group.return_value = defer.fail(UnknownError())
        de = coord.get_coordinator_broker()
        self.successResultOf(de)

    def test_get_coordinator_fatal(self):
        """
        A non-retriable failure when retrieving the coordinator broker leaves
        the coordinator in a quiescent state.
        """
        client = self.mock_client([])
        client._get_coordinator_for_group.return_value = defer.fail(AttributeError())
        coord = self.make_coordinator(client)

        self.failureResultOf(coord.get_coordinator_broker()).check(AttributeError)

        # No heartbeat scheduled.
        self.assertEqual([], client.reactor.getDelayedCalls())

    def test_rejoin_after_error(self):
        """
        try out all the rejoin_after_error scenarios
        """
        client = self.mock_client([])
        coord = self.make_coordinator(client)
        coord.on_group_leave = Mock()

        def check(rejoin_needed, exc):
            coord._rejoin_needed = False
            coord._rejoin_wait_dc = None
            for call in client.reactor.getDelayedCalls():
                call.cancel()
            client.reset_consumer_group_metadata.reset_mock()
            coord.on_group_leave.reset_mock()

            coord.rejoin_after_error(Failure(exc))
            if rejoin_needed:
                self.assertEqual(coord._rejoin_needed, True)
                assert_delayed_calls(1, client)
            else:
                self.assertEqual(coord._rejoin_needed, False)
                assert_delayed_calls(0, client)
                self.assertEqual(coord._rejoin_wait_dc, None)

        check(True, RebalanceInProgress())
        check(True, CoordinatorNotAvailable())
        client.reset_consumer_group_metadata.assert_any_call(coord.group_id)
        check(True, IllegalGeneration())
        coord.on_group_leave.assert_any_call()
        check(True, InvalidGroupId())
        coord.on_group_leave.assert_any_call()
        check(True, InconsistentGroupProtocol())
        check(True, RequestTimedOutError())
        coord.on_group_leave.assert_any_call()
        check(True, UnknownError())

        coord._stopping = True
        check(False, defer.CancelledError())
        coord._stopping = False

        start_d = coord.start()
        start_d.addErrback(lambda f: None)
        check(False, ValueError())
        coord.on_group_leave.assert_any_call()
        self.successResultOf(start_d)


class ConsumerProtocolTests(unittest.SynchronousTestCase):
    """
    Test `afkak._group._ConsumerProtocol`
    """
    def run_protocol(self, topic_partitions, member_subscriptions):
        """
        Run a single round of the group assignment protocol.

        :param topic_partitions:
            The topic partition metadata as loaded from the broker, a map of
            topic name to partition IDs.
        :type topic_partitions: Mapping[str, List[int]]

        :param members: Map of member ID to subscribed topics
        :type members: Mapping[str, List[str]]

        :returns:
            Map of member ID to topic/partition assignments.
        :rtype: Mapping[str, Mapping[str, List[int]]]
        """
        proto = _ConsumerProtocol()
        # Phase 1: Join group. All members join the group.
        joins = {
            member_id: proto.join_group_protocols(topics)[0]
            for member_id, topics in member_subscriptions.items()
        }
        # Phase 2: Sync group. Leader generates assignments.
        sync_assignments = proto.generate_assignments(
            members=[
                _SyncGroupRequestMember(member_id, joins[member_id].protocol_metadata)
                for member_id in member_subscriptions
            ],
            topic_partitions=topic_partitions,
        )
        # Phase 3: Receive assignments.
        return {
            sync_assign.member_id: proto.decode_assignment(sync_assign.member_metadata)
            for sync_assign in sync_assignments
        }

    def test_join_group_protocols(self):
        proto = _ConsumerProtocol()
        self.assertEqual(
            proto.join_group_protocols(['topic1']),
            [_JoinGroupRequestProtocol(
                "consumer",
                b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x00',
            )],
        )

    def test_single_member(self):
        """
        When there is a single group member it is assigned all of the
        partitions.
        """
        proto = _ConsumerProtocol()

        [join_req] = proto.join_group_protocols(['topic1'])
        [sync_assign] = proto.generate_assignments(
            members=[_SyncGroupRequestMember('member_id', join_req.protocol_metadata)],
            topic_partitions={'topic1': [0, 1, 2, 3]},
        )
        assignment = proto.decode_assignment(sync_assign.member_metadata)

        self.assertEqual({"topic1": (0, 1, 2, 3)}, assignment)

    def test_three_members(self):
        """
        Partitions are distributed evenly among members in a round-robin
        fashion.
        """
        assignments = self.run_protocol(
            topic_partitions={'topic2': [0, 1, 2, 3, 4, 5]},
            member_subscriptions={
                'member1': ['topic2'],
                'member2': ['topic2'],
                'member3': ['topic2'],
            },
        )
        self.assertEqual({
            'member1': {'topic2': (0, 3)},
            'member2': {'topic2': (1, 4)},
            'member3': {'topic2': (2, 5)},
        }, assignments)

    def test_roundrobin_normal(self):
        assignments = self.run_protocol(
            topic_partitions={"topic1": [0, 1, 2, 3, 4]},
            member_subscriptions={
                "m1": ["topic1"],
                "m2": ["topic1"],
                "m3": ["topic1"],
            },
        )
        self.assertEqual({
            "m1": {"topic1": (0, 3)},
            "m2": {"topic1": (1, 4)},
            "m3": {"topic1": (2,)},
        }, assignments)

    def test_roundrobin_no_topic(self):
        """
        The protocol refuses to generate empty group assignments. Instead it
        raises a `_NeedTopicPartitions` exception that specifies the required
        topic metatdata.
        """
        with self.assertRaises(_NeedTopicPartitions) as cm:
            self.run_protocol(
                member_subscriptions={
                    "m1": ["topic1"],
                    "m2": ["topic1", "topic2"],
                },
                topic_partitions={},
            )

        self.assertEqual({"topic1", "topic2"}, cm.exception.topics)

    def test_roundrobin_leftover(self):
        """
        If there aren't enough partitions to go around some members may get an
        empty assignment.
        """
        assignments = self.run_protocol(
            topic_partitions={"topic1": [0]},
            member_subscriptions={
                "m1": ["topic1"],
                "m2": ["topic1"],
            },
        )
        self.assertEqual(assignments, {
            "m1": {"topic1": (0,)},
            "m2": {},
        })

    def test_roundrobin_two_topic(self):
        """
        Group members only receive partitions from topics that they have
        subscribed to.
        """
        assignments = self.run_protocol(
            topic_partitions={"topic1": [0], "topic2": [0, 1]},
            member_subscriptions={
                "m1": ["topic1"],
                "m2": ["topic2"],
            },
        )
        self.assertEqual(assignments, {
            "m1": {"topic1": (0,)},
            "m2": {"topic2": (0, 1)},
        })


class TestConsumerGroup(Base):
    def test_repr(self):
        """
        ConsumerGroup's repr displays the group name, instance ID (so that
        instances can be differentiated), state, and any member ID assigned by
        the broker.
        """
        group = ConsumerGroup(self.mock_client([]), "group_id", ["topic1"], lambda c, m: None)
        self.assertRegex(
            repr(group),
            r"\A<afkak\.ConsumerGroup 0x[a-f0-9]+ for 'group_id' \[initialized\] member_id=''>\Z",
        )

    def test_start_stop(self):
        """
            start a consumergroup, join, and start consumers
        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        group.start()
        group.on_join_prepare()
        group.on_join_complete({"topic1": [1, 2, 3]})
        self.assertEqual(len(group.consumers["topic1"]), 3)
        group.stop()
        self.assertEqual(len(group.consumers), 0)

    def test_start_leave(self):
        """
            start a consumergroup, join, start consumers, then get kicked out
        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        group.start()
        group.on_join_prepare()
        group.on_join_complete({"topic1": [1, 2, 3]})
        self.assertEqual(len(group.consumers["topic1"]), 3)
        group.on_group_leave()
        self.assertEqual(len(group.consumers), 0)

    def test_shutdown_error(self):
        """
            get errors while shutting down consumers
        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        group.start()
        with patch('afkak._group.Consumer', side_effect=[Mock(), Mock()]):
            group.on_join_complete({"topic1": [1, 2]})
            consumer = group.consumers["topic1"][0]
            consumer._start_d = defer.Deferred()
            consumer.shutdown.side_effect = KeyError()
            consumer.stop.side_effect = KeyError()
            consumer2 = group.consumers["topic1"][1]
            consumer2.shutdown.return_value = defer.Deferred()

        de = group.shutdown_consumers()
        self.assertNoResult(de)
        self.assertEqual(len(group.consumers), 0)

        consumer2.shutdown.return_value.errback(KeyError())
        consumer2.stop.assert_called_once_with()
        self.successResultOf(de)

    def test_stop_error(self):
        """
            get errors while stopping consumers
        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        group.start()
        with patch('afkak._group.Consumer'):
            group.on_join_complete({"topic1": [1]})
            consumer = group.consumers["topic1"][0]
            consumer.stop.side_effect = KeyError()
        group.stop_consumers()

    def test_consumer_error(self):
        """
            get an unexpected stop error from a consumer
        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        start_d = group.start()
        self.assertNoResult(start_d)
        with patch('afkak._group.Consumer') as mock_consumer:
            mock_consumer.return_value.start.return_value = d = defer.Deferred()
            group.on_join_complete({"topic1": [1]})
            self.assertEqual(mock_consumer.return_value.start.called, True)
            d.errback(Failure(AssertionError()))
            self.failureResultOf(start_d, AssertionError)
            d.addErrback(lambda result: None)

    def test_consumer_cancel_during_shutdown(self):
        """
            get an unexpected CancelledError on the start() deferred
            while shutting down a consumer becasue our heartbeat timed out

        """
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        start_d = group.start()
        with patch('afkak._group.Consumer') as mock_consumer:
            consumer_instance = mock_consumer.return_value
            consumer_start_d = defer.Deferred()
            consumer_instance.start.return_value = consumer_start_d
            consumer_instance._start_d = consumer_start_d
            group.on_join_complete({"topic1": [1]})
            self.assertEqual(consumer_instance.start.called, True)

            def stop():
                consumer_start_d.errback(defer.CancelledError())

            consumer_instance.stop.side_effect = stop
            group.rejoin_after_error(Failure(RequestTimedOutError()))

            self.assertEqual(consumer_instance.stop.called, True)
            self.successResultOf(consumer_start_d)
            self.assertNoResult(start_d)

    def test_rejoin_consumer(self):
        client = self.mock_client([])
        processor = Mock()
        group = ConsumerGroup(client, "group_id", "topic1", processor)
        start_d = group.start()
        group.on_group_leave = Mock()
        with patch('afkak._group.Consumer') as mock_consumer:
            mock_consumer.return_value.start.return_value = d = defer.Deferred()
            group.on_join_complete({"topic1": [1]})
            self.assertEqual(mock_consumer.return_value.start.called, True)
            d.errback(Failure(IllegalGeneration()))
            self.assertEqual(group._rejoin_needed, True)
            self.assertNoResult(start_d)
            group.on_group_leave.assert_any_call()



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the afkak.common module.
"""

from __future__ import absolute_import, division

import unittest

from .. import CODEC_GZIP, CODEC_LZ4, CODEC_SNAPPY, common
from ..common import (
    BrokerResponseError, ConsumerMetadataResponse, CoordinatorLoadInProgress,
    CoordinatorNotAvailable, FetchResponse, LeaderNotAvailableError, Message,
    MessageSizeTooLargeError, NotCoordinator, OffsetCommitResponse,
    OffsetFetchResponse, OffsetMetadataTooLargeError, OffsetOutOfRangeError,
    OffsetResponse, ProduceResponse, RetriableBrokerResponseError,
    UnknownTopicOrPartitionError,
)


class BrokerResponseErrorTests(unittest.TestCase):
    maxDiff = None

    def test_error_codes(self):
        """
        The `afkak.common.BrokerResponseError.errnos` mapping includes all
        concrete subclasses of `BrokerResponseError` by errno attribute.
        """
        count = 0
        expected = {}
        for name in dir(common):
            value = getattr(common, name)
            if not isinstance(value, type):
                continue
            if value is BrokerResponseError:
                continue
            if value is RetriableBrokerResponseError:
                continue
            if value.__name__ != name:
                continue  # Skip backwards-compatibility aliases.
            if issubclass(value, BrokerResponseError):
                expected[value.errno] = value
                count += 1

        self.assertEqual(expected, BrokerResponseError.errnos)
        self.assertEqual(count, len(BrokerResponseError.errnos), "errno values are reused")

    def test_raise_for_errno_ok(self):
        """
        The errno 0 represents "no error", so
        `BrokerResponseError.raise_for_errno()` doesn't raise an exception.
        """
        BrokerResponseError.raise_for_errno(0)

    def test_raise_for_errno_identity(self):
        """
        `BrokerResponseError.raise_for_errno()` raises an exception based on
        the errnos mapping.
        """
        for code, e in BrokerResponseError.errnos.items():
            self.assertRaises(e, BrokerResponseError.raise_for_errno, code)

    def test_raise_for_responses(self):
        """
        `BrokerResponseError.raise_for_errno()` raises the appropriate
        exception for several response types.
        """
        # response (with .error attribute) -> error type
        responses = [
            (ProduceResponse("topic1", 5, 3, 9), UnknownTopicOrPartitionError),
            (FetchResponse("topic2", 3, 10, 8, []), MessageSizeTooLargeError),
            (OffsetResponse("topic3", 8, 1, []), OffsetOutOfRangeError),
            (OffsetCommitResponse("topic4", 10, 12), OffsetMetadataTooLargeError),
            (OffsetFetchResponse("topic5", 33, 12, "", 5), LeaderNotAvailableError),
            (ConsumerMetadataResponse(15, -1, "", -1), CoordinatorNotAvailable),
            (OffsetFetchResponse("topic6", 23, -1, "", 14), CoordinatorLoadInProgress),
            (OffsetCommitResponse("topic7", 24, 16), NotCoordinator),
        ]

        for resp, e in responses:
            self.assertRaises(e, BrokerResponseError.raise_for_errno, resp.error)

    def test_raise_unknown_errno(self):
        """
        `BrokerResponseError` is raised when there is no subclass of it defined
        for a given errno. This exception indicates that the request cannot be
        retried, as we have no information on it.
        """
        unknown_errno = max(BrokerResponseError.errnos.keys()) + 1

        with self.assertRaises(BrokerResponseError) as cm:
            BrokerResponseError.raise_for_errno(unknown_errno)

        self.assertIs(BrokerResponseError, type(cm.exception))  # Not a subclass.
        self.assertEqual(unknown_errno, cm.exception.errno)
        self.assertFalse(cm.exception.retriable)


class MessageTests(unittest.TestCase):
    def test_repr_keyed(self):
        """
        The key of a message is displayed when present.
        """
        self.assertIn(
            repr(Message(0, CODEC_GZIP, b'key', b'')),
            (
                "<Message v0 CODEC_GZIP key='key' value=''>",  # Python 2
                "<Message v0 CODEC_GZIP key=b'key' value=b''>",  # Python 3
            ),
        )

    def test_repr_unkeyed(self):
        """
        A null message key is elided from the message repr.
        """
        self.assertEqual(
            repr(Message(0, 0, None, None)),
            "<Message v0 value=None>",
        )

    def test_repr_compress(self):
        """
        The type of compression is displayed as a keyword.
        """
        self.assertEqual("<Message v0 CODEC_GZIP value=None>", repr(Message(0, CODEC_GZIP, None, None)))
        self.assertEqual("<Message v0 CODEC_SNAPPY value=None>", repr(Message(0, CODEC_SNAPPY, None, None)))
        self.assertEqual("<Message v0 CODEC_LZ4 value=None>", repr(Message(0, CODEC_LZ4, None, None)))

    def test_repr_long_value(self):
        """
        The value of the message is truncated if it exceeds 1024 bytes.
        """
        value = b'a' * 512 + b'b' * 1024
        self.assertEqual(
            '<Message v0 value=1,536 bytes {!r}...>'.format(b'a' * 512),
            repr(Message(0, 0, None, value)),
        )
